"""FFT Through Focus MTF analysis."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Annotated, Literal, Union

import numpy as np
import pandas as pd
from pydantic import Field, RootModel

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.types import UnitField, ValidatedDataFrame, ZOSAPIConstant
from zospy.api import config, constants
from zospy.utils.pyutils import atox
from zospy.utils.zputils import standardize_sampling

if TYPE_CHECKING:
    from collections.abc import Iterator

__all__ = ("FFTThroughFocusMTF", "FFTThroughFocusMTFSettings")


@analysis_settings
class FFTThroughFocusMTFSettings:
    """Settings for the FFT Through Focus MTF analysis.

    For an in depth explanation of the parameters, see the OpticStudio user manual.

    Attributes
    ----------
    sampling : str | int
        The pupil sampling, either string (e.g. "64x64") or int. Integers will be treated as a ZOSAPI Constants
        integer and must be greater than or equal to 0. Defaults to "64x64".
    delta_focus : float
        The Z-axis range or optical power range, depending on the optical system. Defaults to 0.1.
    frequency : float
        The spatial frequency for which the MTF is calculated. Defaults to 0.
    number_of_steps : int
        The number of steps in the focus range. Defaults to 5.
    wavelength : int | str
        The wavelength to use in the MTF. Either 'All' or an integer specifying the wavelength number.
    field : str | int
        The field to use in the MTF. Either 'All' or an integer specifying the field number.
    mtf_type : zospy.constants.Analysis.Settings.Mtf.MtfTypes.Modulation
        The MTF type (e.g. `Modulation`) that is calculated.
    use_polarization : bool
        Use polarization. Defaults to False.
    use_dashes : bool
        Use dashes. Defaults to False.
    """

    sampling: str | Annotated[int, Field(ge=0)] = Field(default="64x64", description="Sampling grid size")
    delta_focus: float = Field(default=0.1, description="Focus step size")
    frequency: float = Field(default=0, description="Frequency")
    number_of_steps: int = Field(default=5, ge=0, description="Number of steps")
    wavelength: Literal["All"] | Annotated[int, Field(gt=0)] = Field(
        default="All", description="Wavelength number or 'All'"
    )
    field: Literal["All"] | Annotated[int, Field(gt=0)] = Field(default="All", description="Field number or 'All'")
    mtf_type: ZOSAPIConstant("Analysis.Settings.Mtf.MtfTypes") = Field(default="Modulation", description="MTF type")
    use_polarization: bool = Field(default=False, description="Use polarization")
    use_dashes: bool = Field(default=False, description="Use dashes")


@analysis_result
class FFTThroughFocusMTFData:
    field_coordinate: UnitField[tuple[float, float]]
    data: ValidatedDataFrame

    def to_dataframe(self) -> pd.DataFrame:
        """Convert the data to a Pandas DataFrame.

        In addition to the columns from FFTThroughFocusMTFData.data, the returned DataFrame has the following columns:

        - FieldX: The field x coordinate
        - FieldY: The field y coordinate

        Returns
        -------
        DataFrame
            The data in long format.
        """
        df: pd.DataFrame = self.data.copy().reset_index()

        df.insert(0, "FieldX", self.field_coordinate.value[0])
        df.insert(1, "FieldY", self.field_coordinate.value[1])

        return df


class FFTThroughFocusMTFResult(RootModel[list[FFTThroughFocusMTFData]]):
    def __iter__(self) -> Iterator[FFTThroughFocusMTFData]:
        return iter(self.root)

    def __getitem__(self, item) -> FFTThroughFocusMTFData:
        return self.root[item]

    def __len__(self):
        return len(self.root)

    def to_dataframe(self) -> pd.DataFrame:
        """Convert the data to a Pandas DataFrame.

        The separate dataframes for each field are combined in a DataFrame in long format.
        In addition to the columns for each wavelength, the returned DataFrame has the following columns:

        - Direction: The direction of the fan, either 'Tangential' or 'Sagittal'.
        - Field Number: The field number.
        - Field: The field coordinate.

        Returns
        -------
        DataFrame
            The data in long format.
        """
        return pd.concat([fft.to_dataframe() for fft in self], ignore_index=True).reset_index(drop=True)


class FFTThroughFocusMTF(
    BaseAnalysisWrapper[Union[FFTThroughFocusMTFResult, None], FFTThroughFocusMTFSettings],
    analysis_type="FftThroughFocusMtf",
    needs_config_file=True,
):
    """FFT Through Focus MTF analysis."""

    def __init__(
        self,
        *,
        sampling: str | int = "64x64",
        delta_focus: float = 0.1,
        frequency: float = 0,
        number_of_steps: int = 5,
        wavelength: int | Literal["All"] = "All",
        field: int | Literal["All"] = "All",
        mtf_type: constants.Analysis.Settings.Mtf.MtfTypes | str = "Modulation",
        use_polarization: bool = False,
        use_dashes: bool = False,
    ):
        """Create a new FFT Through Focus MTF analysis.

        See Also
        --------
        FFTThroughFocusMTFSettings : Settings for the FFT Through Focus MTF analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> FFTThroughFocusMTFResult | None:
        """Run the FFT Through Focus MTF analysis."""
        self.analysis.Settings.SampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.sampling)
        )
        self.analysis.Settings.DeltaFocus = self.settings.delta_focus
        self.analysis.Settings.Frequency = self.settings.frequency
        self.analysis.Settings.NumberOfSteps = self.settings.number_of_steps
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.Type = constants.process_constant(
            constants.Analysis.Settings.Mtf.MtfTypes, self.settings.mtf_type
        )
        self.analysis.Settings.UsePolarization = self.settings.use_polarization
        self.analysis.Settings.UseDashes = self.settings.use_dashes

        self._correct_mtf_type_api_bug()

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.get_data_series()

    def get_data_series(self) -> FFTThroughFocusMTFResult | None:
        """Get the data series from the FFT Through Focus MTF analysis."""
        re_float = rf"-?\d+\{config.DECIMAL_POINT}\d+"
        fft_through_focus_mtf_description_regex = re.compile(
            rf"Field: "
            rf"(?P<field_1>{re_float})(?:, (?P<field_2>{re_float}))? "
            r"(?P<unit>\(.+?\)|\S.*)?",  # unit can, but might not have parentheses
            re.IGNORECASE,
        )

        if self.analysis.Results.NumberOfDataSeries <= 0:
            return None

        fft_results = []
        for i in range(self.analysis.Results.NumberOfDataSeries):
            data_series = self.analysis.Results.GetDataSeries(i)

            match = fft_through_focus_mtf_description_regex.match(data_series.Description)

            if match is None:
                raise ValueError(f"Could not parse description: {data_series.Description}")

            index = pd.Index(data_series.XData.Data, name=data_series.XLabel)
            columns = data_series.SeriesLabels
            data = np.array(data_series.YData.Data)

            if match.group("field_2"):
                field_x = atox(match.group("field_1"), float)
                field_y = atox(match.group("field_2"), float)
            else:  # field 1 corresponds to y
                field_x = 0.0
                field_y = atox(match.group("field_1"), float)

            coordinate = (field_x, field_y)

            fft_data = FFTThroughFocusMTFData(
                field_coordinate=UnitField(value=coordinate, unit=match.group("unit")),
                data=pd.DataFrame(index=index, columns=columns, data=data),
            )

            fft_results.append(fft_data)

        return FFTThroughFocusMTFResult.model_validate(fft_results)

    def _correct_mtf_type_api_bug(self) -> None:
        """Correction for an API bug in OpticStudio versions < 21.2.

        In these versions, the MTF Type cannot be set through the ZOS-API for the FFT Through Focus MTF
        See also: https://community.zemax.com/zos-api-12/zos-api-setting-mtf-property-type-not-working-730

        Returns
        -------
        None
        """
        if self.oss.ZOS.version < "21.2.0":
            config_file = str(self.config_file)

            self.analysis.Settings.SaveTo(config_file)

            self.analysis.Settings.ModifySettings(
                config_file,
                "TFM_TYPE",
                str(int(constants.process_constant(constants.Analysis.Settings.Mtf.MtfTypes, self.settings.mtf_type))),
            )
            self.analysis.Settings.LoadFrom(config_file)



import pytest
from pandas.testing import assert_frame_equal

from tests.helpers import assert_dataclass_equal
from zospy.analyses.raysandspots import RayFan, SingleRayTrace


class TestSingleRayTrace:
    def test_can_run(self, simple_system):
        result = SingleRayTrace().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = SingleRayTrace().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "hx,hy,px,py,raytrace_type,global_coordinates",
        [
            (1, 1, 0, 0, "DirectionCosines", False),
            (0, 0, 1, 1, "DirectionCosines", True),
            (1, 1, 0, 0, "TangentAngle", True),
        ],
    )
    def test_single_ray_trace_returns_correct_result(
        self, simple_system, hx, hy, px, py, raytrace_type, global_coordinates, expected_data
    ):
        result = SingleRayTrace(
            hx=hx,
            hy=hy,
            px=px,
            py=py,
            raytrace_type=raytrace_type,
            global_coordinates=global_coordinates,
        ).run(simple_system)

        assert_dataclass_equal(result.data, expected_data.data)

    @pytest.mark.parametrize(
        "hx,hy,px,py,raytrace_type,global_coordinates",
        [
            (1, 1, 0, 0, "DirectionCosines", False),
            (0, 0, 1, 1, "DirectionCosines", True),
            (1, 1, 0, 0, "TangentAngle", True),
        ],
    )
    def test_single_ray_trace_matches_reference_data(
        self, simple_system, hx, hy, px, py, raytrace_type, global_coordinates, reference_data
    ):
        result = SingleRayTrace(
            hx=hx,
            hy=hy,
            px=px,
            py=py,
            raytrace_type=raytrace_type,
            global_coordinates=global_coordinates,
        ).run(simple_system)

        assert_dataclass_equal(
            result.data, reference_data.data, ignore_fields=["real_ray_trace_data", "paraxial_ray_trace_data"]
        )


class TestRayFan:
    def test_can_run(self, simple_system):
        result = RayFan().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = RayFan().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "fieldx,fieldy", [(0, 0), (5.5, 0), (0, 5.5), (5.5, 5.5), (-5.5, 0), (0, -5.5), (-5.5, -5.5)]
    )
    def test_field_parsing(self, fieldx, fieldy, simple_system):
        field1 = simple_system.SystemData.Fields.GetField(1)
        field1.X = fieldx
        field1.Y = fieldy
        result = RayFan().run(simple_system)

        assert result.data.tangential[0].field_coordinate.value[0] == fieldx
        assert result.data.tangential[0].field_coordinate.value[1] == fieldy
        assert result.data.sagittal[0].field_coordinate.value[0] == fieldx
        assert result.data.sagittal[0].field_coordinate.value[1] == fieldy

    @pytest.mark.parametrize("fields", [[(5.5, -5.5)], [(0, 0), (0.0, 5.5), (5.5, 0.0), (5.5, -5.5)]])
    def test_to_dataframe(self, fields, simple_system):
        field1 = simple_system.SystemData.Fields.GetField(1)
        field1.X = fields[0][0]
        field1.Y = fields[0][1]

        for f in fields[1:]:
            simple_system.SystemData.Fields.AddField(f[0], f[1], 1.0)

        result = RayFan().run(simple_system)

        df = result.data.to_dataframe()

        for r in result.data.tangential:
            for wl in r.data.columns:
                assert_frame_equal(
                    df.loc[
                        (df["Direction"] == "Tangential")
                        & (df["Field Number"] == r.field_number)
                        & (df["FieldX"] == r.field_coordinate.value[0])
                        & (df["FieldY"] == r.field_coordinate.value[1])
                        & (df["Wavelength"] == wl),
                        ["Pupil", "Aberration"],
                    ].set_index("Pupil"),
                    r.data[wl].to_frame("Aberration"),
                )

        for r in result.data.sagittal:
            for wl in r.data.columns:
                assert_frame_equal(
                    df.loc[
                        (df["Direction"] == "Sagittal")
                        & (df["Field Number"] == r.field_number)
                        & (df["FieldX"] == r.field_coordinate.value[0])
                        & (df["FieldY"] == r.field_coordinate.value[1])
                        & (df["Wavelength"] == wl),
                        ["Pupil", "Aberration"],
                    ].set_index("Pupil"),
                    r.data[wl].to_frame("Aberration"),
                )

    @pytest.mark.parametrize(
        "plot_scale,number_of_rays,tangential,sagittal",
        [
            (0, 20, "Aberration_Y", "Aberration_X"),
            (1, 40, "Aberration_Y", "Aberration_X"),
            (0, 20, "Aberration_X", "Aberration_Y"),
        ],
    )
    def test_ray_fan_returns_correct_result(
        self, simple_system, plot_scale, number_of_rays, tangential, sagittal, expected_data
    ):
        result = RayFan(
            plot_scale=plot_scale,
            number_of_rays=number_of_rays,
            tangential=tangential,
            sagittal=sagittal,
        ).run(simple_system)

        assert_dataclass_equal(result.data, expected_data.data)

    @pytest.mark.parametrize(
        "plot_scale,number_of_rays,tangential,sagittal",
        [
            (0, 20, "Aberration_Y", "Aberration_X"),
            (1, 40, "Aberration_Y", "Aberration_X"),
            (0, 20, "Aberration_X", "Aberration_Y"),
        ],
    )
    def test_ray_fan_matches_reference_data(
        self, simple_system, plot_scale, number_of_rays, tangential, sagittal, reference_data
    ):
        result = RayFan(
            plot_scale=plot_scale,
            number_of_rays=number_of_rays,
            tangential=tangential,
            sagittal=sagittal,
        ).run(simple_system)

        assert_dataclass_equal(result.data, reference_data.data)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "BackgroundOptions",
    "BrightnessOptions",
    "CameraViewpointOptions",
    "ColorRaysByCrossSectionOptions",
    "ColorRaysByNSCOptions",
    "ColorRaysByOptions",
    "DetectorDisplayModeOptions",
    "DetectorPixelColorOptions",
    "DrawSectionOptions",
    "LineThicknessOptions",
    "NumberSegmentsOptions",
    "OpacityOptions",
    "RayTraceOptions",
    "RealPupilOptions",
)

class BackgroundOptions:
    White: BackgroundOptions = None
    Black: BackgroundOptions = None
    Red: BackgroundOptions = None
    Green: BackgroundOptions = None
    Blue: BackgroundOptions = None
    DarkGreen: BackgroundOptions = None
    DarkBlue: BackgroundOptions = None
    Color01: BackgroundOptions = None
    Color02: BackgroundOptions = None
    Color03: BackgroundOptions = None
    Color04: BackgroundOptions = None
    Color05: BackgroundOptions = None
    Color06: BackgroundOptions = None
    Color07: BackgroundOptions = None
    Color08: BackgroundOptions = None
    Color09: BackgroundOptions = None
    Color10: BackgroundOptions = None
    Color11: BackgroundOptions = None
    Color12: BackgroundOptions = None
    Color13: BackgroundOptions = None
    Color14: BackgroundOptions = None
    Color15: BackgroundOptions = None
    Color16: BackgroundOptions = None
    Color17: BackgroundOptions = None
    Color18: BackgroundOptions = None
    Color19: BackgroundOptions = None
    Color20: BackgroundOptions = None
    Color21: BackgroundOptions = None
    Color22: BackgroundOptions = None
    Color23: BackgroundOptions = None
    Color24: BackgroundOptions = None
    Gradient01: BackgroundOptions = None
    Gradient02: BackgroundOptions = None
    Gradient03: BackgroundOptions = None
    Gradient04: BackgroundOptions = None
    Gradient05: BackgroundOptions = None
    Gradient06: BackgroundOptions = None
    Gradient07: BackgroundOptions = None
    Gradient08: BackgroundOptions = None
    Gradient09: BackgroundOptions = None
    Gradient10: BackgroundOptions = None
    GradientUser: BackgroundOptions = None

class BrightnessOptions:
    P100: BrightnessOptions = None
    P90: BrightnessOptions = None
    P80: BrightnessOptions = None
    P70: BrightnessOptions = None
    P60: BrightnessOptions = None
    P50: BrightnessOptions = None
    P40: BrightnessOptions = None
    P30: BrightnessOptions = None
    P20: BrightnessOptions = None
    P10: BrightnessOptions = None

class CameraViewpointOptions:
    Isometric: CameraViewpointOptions = None
    XY: CameraViewpointOptions = None
    YZ: CameraViewpointOptions = None
    XZ: CameraViewpointOptions = None

class ColorRaysByCrossSectionOptions:
    Fields: ColorRaysByCrossSectionOptions = None
    Waves: ColorRaysByCrossSectionOptions = None
    Wavelength: ColorRaysByCrossSectionOptions = None

class ColorRaysByNSCOptions:
    SourceNumber: ColorRaysByNSCOptions = None
    WaveNumber: ColorRaysByNSCOptions = None
    ConfigNumber: ColorRaysByNSCOptions = None
    Wavelength: ColorRaysByNSCOptions = None
    SegmentNumber: ColorRaysByNSCOptions = None

class ColorRaysByOptions:
    Fields: ColorRaysByOptions = None
    Waves: ColorRaysByOptions = None
    Config: ColorRaysByOptions = None
    Wavelength: ColorRaysByOptions = None

class DetectorDisplayModeOptions:
    Consider: DetectorDisplayModeOptions = None
    GreyScaleFlux: DetectorDisplayModeOptions = None
    InverseGreyScaleFlux: DetectorDisplayModeOptions = None
    FalseColorFlux: DetectorDisplayModeOptions = None
    InverseFalseColorFlux: DetectorDisplayModeOptions = None
    GreyScaleIrradiance: DetectorDisplayModeOptions = None
    InverseGreyScaleIrradiance: DetectorDisplayModeOptions = None
    FalseColorIrradiance: DetectorDisplayModeOptions = None
    InverseFalseColorIrradiance: DetectorDisplayModeOptions = None

class DetectorPixelColorOptions:
    DoNotColorIndividualPoints: DetectorPixelColorOptions = None
    ByRaysOnLayout: DetectorPixelColorOptions = None
    ByLastAnalysis: DetectorPixelColorOptions = None

class DrawSectionOptions:
    P100: DrawSectionOptions = None
    P75: DrawSectionOptions = None
    P50: DrawSectionOptions = None
    P25: DrawSectionOptions = None

class LineThicknessOptions:
    Standard: LineThicknessOptions = None
    Thick: LineThicknessOptions = None
    Thickest: LineThicknessOptions = None
    Thinnest: LineThicknessOptions = None
    Thin: LineThicknessOptions = None

class NumberSegmentsOptions:
    S_8: NumberSegmentsOptions = None
    S_16: NumberSegmentsOptions = None
    S_32: NumberSegmentsOptions = None
    S_64: NumberSegmentsOptions = None
    S_128: NumberSegmentsOptions = None

class OpacityOptions:
    Ignore: OpacityOptions = None
    Consider: OpacityOptions = None
    All50Percent: OpacityOptions = None

class RayTraceOptions:
    UseRays: RayTraceOptions = None
    LightningTraceTrueColor: RayTraceOptions = None
    LightningTraceAvgWavelength: RayTraceOptions = None

class RealPupilOptions:
    Pupils_Off: RealPupilOptions = None
    Pupils_4: RealPupilOptions = None
    Pupils_8: RealPupilOptions = None
    Pupils_16: RealPupilOptions = None
    Pupils_32: RealPupilOptions = None



import pytest

import zospy as zp
from zospy import solvers
from zospy.api.config import DECIMAL_POINT


def test_element_power(simple_system):
    lens_back = simple_system.LDE.GetSurfaceAt(3)

    solvers.element_power(lens_back.RadiusCell, power=40)

    assert lens_back.RadiusCell.GetSolveData().Type == zp.constants.Editors.SolveType.ElementPower
    assert lens_back.Radius != -20


def test_fixed(simple_system):
    lens_back = simple_system.LDE.GetSurfaceAt(3)

    solvers.fixed(lens_back.RadiusCell)

    assert lens_back.RadiusCell.GetSolveData().Type == zp.constants.Editors.SolveType.Fixed


def test_material_model(simple_system):
    lens_front = simple_system.LDE.GetSurfaceAt(2)

    solvers.material_model(lens_front.MaterialCell, refractive_index=2.0, abbe_number=1, partial_dispersion=0)

    assert lens_front.MaterialCell.GetSolveData().Type == zp.constants.Editors.SolveType.MaterialModel
    assert lens_front.Material == f"2{DECIMAL_POINT}00,1,0"


@pytest.mark.parametrize("from_column,scale,offset", [(None, 1, 0), ("Radius", 2, 1)])
def test_surface_pickup(simple_system, from_column, scale, offset):
    lens_front = simple_system.LDE.GetSurfaceAt(2)
    lens_back = simple_system.LDE.GetSurfaceAt(3)

    solvers.surface_pickup(lens_back.ThicknessCell, lens_front, from_column, scale, offset)

    assert lens_back.ThicknessCell.GetSolveData().Type == zp.constants.Editors.SolveType.SurfacePickup

    if from_column is None:
        assert lens_back.Thickness == lens_front.Thickness
    else:
        assert (
            lens_back.Thickness
            == lens_front.GetSurfaceCell(getattr(zp.constants.Editors.LDE.SurfaceColumn, from_column)).DoubleValue
            * scale
            + offset
        )


def test_pickup_chief_ray(coordinate_break_system):
    coordinate_break = coordinate_break_system.LDE.GetSurfaceAt(4)

    solvers.pickup_chief_ray(coordinate_break.SurfaceData.Decenter_X_Cell)

    assert (
        coordinate_break.SurfaceData.Decenter_X_Cell.GetSolveData().Type
        == zp.constants.Editors.SolveType.PickupChiefRay
    )
    assert coordinate_break.SurfaceData.Decenter_X <= 1e-3


def test_position(simple_system):
    length = 50

    stop = simple_system.LDE.GetSurfaceAt(2)
    lens_back = simple_system.LDE.GetSurfaceAt(3)

    solvers.position(lens_back.ThicknessCell, from_surface=stop, length=length)

    assert lens_back.ThicknessCell.GetSolveData().Type == zp.constants.Editors.SolveType.Position
    assert sum(simple_system.LDE.GetSurfaceAt(i).Thickness for i in range(1, 4)) == length


def test_variable(simple_system):
    lens_back = simple_system.LDE.GetSurfaceAt(3)

    solvers.variable(lens_back.RadiusCell)

    assert lens_back.RadiusCell.GetSolveData().Type == zp.constants.Editors.SolveType.Variable



"""OpticStudio analyses from the Physical Optics category."""

from zospy.analyses.physicaloptics.physical_optics_propagation import (
    PhysicalOpticsPropagation,
    PhysicalOpticsPropagationSettings,
    create_beam_parameter_dict,
    create_fiber_parameter_dict,
)

__all__ = (
    "PhysicalOpticsPropagation",
    "PhysicalOpticsPropagationSettings",
    "create_beam_parameter_dict",
    "create_fiber_parameter_dict",
)



"""Utility functions for working with Python types."""

from __future__ import annotations

from operator import attrgetter
from pathlib import Path
from sys import version_info
from typing import TYPE_CHECKING, TypeVar

import zospy.api.config as _config

if TYPE_CHECKING:
    from collections.abc import Callable
    from os import PathLike


def _check_path(path: Path, *, directory_only: bool = False) -> bool:
    if directory_only and not path.is_dir():
        return _check_path(path.parent)

    return path.exists()


def abspath(path: PathLike | str, *, check_directory_only: bool = False) -> str:
    """Convert a path to an absolute path and check if it exists.

    Parameters
    ----------
    path : Path | str
        The path to be made absolute.
    check_directory_only : bool
        Whether to check only if directories exist. If set to `True`, parent directories are verified to exist, but
        files are not.

    Returns
    -------
    str
        The absolute path, converted to a string.

    Raises
    ------
    FileNotFoundError
        If `check_directory_only` is False and the path does not exist, or if `check_directory_only` is True and the
        parent directory does not exist.
    """
    # Behaviour of Path.resolve when the file does not exist is incorrect prior to Python 3.10
    absolute_path = Path(path).absolute() if version_info <= (3, 10) else Path(path).resolve()

    if _check_path(absolute_path, directory_only=check_directory_only):
        return str(absolute_path)
    raise FileNotFoundError(absolute_path)


def attrsetter(obj, attr, val):
    """Set an attribute of an object.

    Parameters
    ----------
    obj
        The object from which the attribute is set
    attr
        The name of the attribute
    val
        The value to which the attribute is set

    Returns
    -------
        None
    """
    pre, _, post = attr.rpartition(".")
    return setattr(attrgetter(pre)(obj) if pre else obj, post, val)


def _delocalize(
    string: str,
    decimal_point: str = ...,
    thousands_separator: str | None = ...,
) -> str:
    """Delocalize a string as a normalized number.

    By default, the locale settings stored in zospy.api.config are used.

    Parameters
    ----------
    string : str
        The string that is to be converted.
    decimal_point : str
        The decimal point separator used in the string. Defaults to zospy.api.config.DECIMAL_POINT.
    thousands_separator : str | None
        The thousands separator used in the string. Defaults to zospy.api.config.THOUSANDS_SEPARATOR.

    Returns
    -------
    str
        The delocalized string.
    """
    # Get rid of the thousands grouping
    thousands_separator = _config.THOUSANDS_SEPARATOR if thousands_separator is ... else thousands_separator
    decimal_point = _config.DECIMAL_POINT if decimal_point is ... else decimal_point

    if thousands_separator:
        string = string.replace(thousands_separator, "")

    # Replace the decimal point with a dot
    if decimal_point:
        string = string.replace(decimal_point, ".")

    return string


Number = TypeVar("Number", int, float)


def atox(
    string: str,
    dtype: Callable[[str], Number] = float,
    decimal_point: str = ...,
    thousands_separator: str | None = ...,
) -> Number:
    """Parse a string to a numeric type.

    By default, the locale settings stored in zospy.api.config are used to delocalize the string.

    Parameters
    ----------
    string : str
        The string that is to be converted.
    dtype : Callable[[str], int | float]
        The function used to convert the delocalized string into a number.
    decimal_point : str
        The decimal point separator used in the string. Defaults to zospy.api.config.DECIMAL_POINT.
    thousands_separator : str | None
        The thousands separator used in the string. Defaults to zospy.api.config.THOUSANDS_SEPARATOR.

    Returns
    -------
    str
        The delocalized string.
    """
    return dtype(_delocalize(string, decimal_point=decimal_point, thousands_separator=thousands_separator))


def xtoa(
    number: float,
    decimal_point: str = ...,
    thousands_separator: str | None = ...,
) -> str:
    """Localize a number back to a string using the locale settings.

    By default, the locale settings stored in zospy.api.config are used.

    Parameters
    ----------
    number : int | float
        The number that is to be converted.
    decimal_point : str
        The decimal point separator used in the string. Defaults to zospy.api.config.DECIMAL_POINT.
    thousands_separator : str | None
        The thousands separator used in the string. Defaults to zospy.api.config.THOUSANDS_SEPARATOR.

    Returns
    -------
    str
        The localized string representation of the number.
    """
    thousands_separator = _config.THOUSANDS_SEPARATOR if thousands_separator is ... else thousands_separator
    decimal_point = _config.DECIMAL_POINT if decimal_point is ... else decimal_point

    if isinstance(number, int):
        return format(number, ",").replace(",", thousands_separator) if thousands_separator else str(number)

    if isinstance(number, float):
        if not decimal_point:
            raise ValueError("Converting float to string requires decimal point to be known")

        string = format(number, ",") if thousands_separator else str(number)

        # swap , and . simultaneously for thousands_separator and decimal_point
        return string.translate(str.maketrans({",": thousands_separator, ".": decimal_point}))

    raise TypeError(f"Expected int or float, got {type(number).__name__}")



from __future__ import annotations

from pathlib import Path
from typing import Literal

import pytest
from semver.version import Version

import zospy as zp


def pytest_addoption(parser):
    parser.addoption("--extension", action="store_true", help="Connect to Zemax OpticStudio as extension")
    parser.addoption("--output-directory", type=Path)
    parser.addoption("--update-ui", type=bool, default=False)
    parser.addoption(
        "--opticstudio-directory", type=Path, default=None, help="Path to the OpticStudio installation directory"
    )
    parser.addoption("--old-analyses", action="store_true", help="Run tests for old analyses")


def pytest_runtest_makereport(item, call):
    if any(m.name == "must_pass" for m in item.iter_markers()) and call.excinfo is not None:
        pytest.exit(f"Aborting because a must pass test failed: {item.name}", 1)


def _get_old_analyses(items):
    return [item for item in items if item.get_closest_marker("old_analyses")]


@pytest.hookimpl(hookwrapper=True)
def pytest_collection_modifyitems(config, items):
    """Customize test selection.

    Customizing test selection allows to hide results from the test output instead of marking them as skipped.
    Current customizations:
    - Skip tests for old analyses if the `--old-analyses` option is not set.
    - Run must_pass tests first.
    """
    yield

    # Deselect tests for old analyses if --old-analyses is not set
    if not config.getoption("--old-analyses"):
        deselected = _get_old_analyses(items)
        items[:] = [item for item in items if item not in deselected]
        config.hook.pytest_deselected(items=deselected)

    # Run must_pass tests first
    must_pass_tests = [item for item in items if item.get_closest_marker("must_pass")]
    for test in must_pass_tests:
        items.remove(test)
        items.insert(0, test)


@pytest.fixture(scope="session")
def connection_mode(request) -> Literal["extension", "standalone"]:
    return "extension" if request.config.getoption("--extension") else "standalone"


@pytest.fixture(autouse=True)
def skip_by_connection_mode(request, connection_mode):
    if request.node.get_closest_marker("require_mode"):
        required_mode = request.node.get_closest_marker("require_mode").args[0]
        if required_mode != connection_mode:
            pytest.skip(f"Test is only applicable in {required_mode} mode.")


@pytest.fixture(autouse=True)
def skip_for_opticstudio_versions(request, optic_studio_version):
    if request.node.get_closest_marker("skip_for_opticstudio_versions"):
        conditions, reason = request.node.get_closest_marker("skip_for_opticstudio_versions").args[:2]

        if isinstance(conditions, str):
            conditions = [conditions]

        for condition in conditions:
            if optic_studio_version.match(condition):
                pytest.skip(reason=reason)


@pytest.fixture(autouse=True)
def xfail_for_opticstudio_versions(request, optic_studio_version):
    if request.node.get_closest_marker("xfail_for_opticstudio_versions"):
        conditions, reason = request.node.get_closest_marker("xfail_for_opticstudio_versions").args[:2]

        if isinstance(conditions, str):
            conditions = [conditions]

        for condition in conditions:
            if optic_studio_version.match(condition):
                request.node.add_marker(pytest.mark.xfail(True, reason=reason, raises=AssertionError))


@pytest.fixture(autouse=True)
def skip_old_analysis_tests(request):
    if request.node.get_closest_marker("old_analyses") and not request.config.getoption("--old-analyses"):
        pytest.skip("Skipping tests for old analyses")


@pytest.fixture(scope="session")
def system_save_file(request):
    output_directory = request.config.getoption("--output-directory")

    if output_directory:
        save_file = output_directory / f"{request.fspath.basename}-{request.node.name}.zos"
        return save_file.absolute()
    return None


@pytest.fixture(scope="session")
def opticstudio_directory(request) -> Path | None:
    path = request.config.getoption("--opticstudio-directory")

    if path is not None:
        return Path(path).resolve(strict=True)

    return None


@pytest.fixture(scope="session")
def zos(opticstudio_directory) -> zp.ZOS:
    return zp.ZOS(opticstudio_directory=str(opticstudio_directory)) if opticstudio_directory is not None else zp.ZOS()


@pytest.fixture
def oss(zos: zp.ZOS, connection_mode, request) -> zp.zpcore.OpticStudioSystem:
    oss = zos.connect(connection_mode)

    if connection_mode == "extension":
        # Disable UI updates using command line option, making the tests run faster
        zos.Application.ShowChangesInUI = request.config.getoption("--update-ui")

    yield oss

    # Close the system
    if zos.Connection.IsAlive:
        zos.Application.CloseApplication()
        zos.Application = None


@pytest.fixture(scope="session")
def optic_studio_version(zos, connection_mode) -> Version:
    zos.connect(connection_mode)

    version = Version(
        major=zos.Application.ZOSMajorVersion, minor=zos.Application.ZOSMinorVersion, patch=zos.Application.ZOSSPVersion
    )

    zos.Application.CloseApplication()
    zos.Application = None

    return version


@pytest.fixture
def empty_system(oss, system_save_file) -> zp.zpcore.OpticStudioSystem:
    oss.new()
    oss.make_sequential()

    yield oss

    if system_save_file:
        oss.save_as(str(system_save_file))


@pytest.fixture
def simple_system(empty_system) -> zp.zpcore.OpticStudioSystem:
    oss = empty_system

    oss.SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.FloatByStopSize
    oss.SystemData.Wavelengths.GetWavelength(1).Wavelength = 0.543

    surface_object = oss.LDE.GetSurfaceAt(0)
    surface_object.Thickness = float("inf")

    surface_stop = oss.LDE.GetSurfaceAt(1)
    surface_stop.SemiDiameter = 1

    lens_front = oss.LDE.InsertNewSurfaceAt(2)
    lens_front.Comment = "lens front"
    lens_front.Radius = 20
    lens_front.Thickness = 1
    zp.solvers.material_model(lens_front.MaterialCell, refractive_index=1.5)

    lens_back = oss.LDE.InsertNewSurfaceAt(3)
    lens_back.Comment = "lens back"
    lens_back.Radius = -20
    lens_back.Thickness = 19.792  # System is in focus

    return oss


@pytest.fixture
def polarized_system(simple_system) -> zp.zpcore.OpticStudioSystem:
    """Simple system with a polarizing front lens surface."""
    oss = simple_system

    lens_front = oss.LDE.GetSurfaceAt(2)

    # Change surface to Jones Matrix
    zp.functions.lde.surface_change_type(lens_front, zp.constants.Editors.LDE.SurfaceType.JonesMatrix)

    # Set Jones Matrix elements to horizontal polarization
    lens_front.GetCellAt(12).DoubleValue = 1  # A real
    lens_front.GetCellAt(18).DoubleValue = 0  # D real

    return oss


@pytest.fixture
def coordinate_break_system(simple_system: zp.zpcore.OpticStudioSystem) -> zp.zpcore.OpticStudioSystem:
    """Simple system with a coordinate break behind the lens."""
    oss = simple_system

    coordinate_break = oss.LDE.InsertNewSurfaceAt(4)
    zp.functions.lde.surface_change_type(coordinate_break, zp.constants.Editors.LDE.SurfaceType.CoordinateBreak)
    coordinate_break.SurfaceData.Decenter_X = 2

    return oss


@pytest.fixture
def decentered_system(simple_system) -> zp.zpcore.OpticStudioSystem:
    """Simple system with incoming rays at a non-zero angle."""
    oss = simple_system

    field = oss.SystemData.Fields.GetField(1)
    field.X = 10
    field.Y = 20

    return oss


@pytest.fixture
def nsc_empty_system(oss, system_save_file) -> zp.zpcore.OpticStudioSystem:
    oss.new()
    oss.make_nonsequential()

    yield oss

    if system_save_file:
        oss.save_as(str(system_save_file))


@pytest.fixture
def nsc_simple_system(nsc_empty_system) -> zp.zpcore.OpticStudioSystem:
    oss = nsc_empty_system

    oss.SystemData.Wavelengths.GetWavelength(1).Wavelength = 0.543

    source_object = oss.NCE.GetObjectAt(1)
    zp.functions.nce.object_change_type(source_object, zp.constants.Editors.NCE.ObjectType.SourceEllipse)
    source_object.GetCellAt(11).IntegerValue = 100  # Layout Rays
    source_object.GetCellAt(12).IntegerValue = 100  # Analysis Rays
    source_object.GetCellAt(16).DoubleValue = 1  # X Half Width
    source_object.GetCellAt(17).DoubleValue = 1  # Y Half Width

    lens_object = oss.NCE.InsertNewObjectAt(2)
    zp.functions.nce.object_change_type(lens_object, zp.constants.Editors.NCE.ObjectType.StandardLens)
    lens_object.ZPosition = 10
    lens_object.GetCellAt(11).DoubleValue = 20  # Radius front
    lens_object.GetCellAt(13).DoubleValue = 2  # Semi diameter front
    lens_object.GetCellAt(15).DoubleValue = 1  # Thickness
    lens_object.GetCellAt(16).DoubleValue = -20
    lens_object.GetCellAt(18).DoubleValue = 2  # Semi diameter back

    zp.solvers.material_model(lens_object.MaterialCell, refractive_index=1.5)

    detector_object = oss.NCE.InsertNewObjectAt(3)
    zp.functions.nce.object_change_type(detector_object, zp.constants.Editors.NCE.ObjectType.DetectorSurface)
    detector_object.ZPosition = 10 + 1 + 19.792
    return oss



"""Configuration module for ZOSPy.

Retrieves and stores the number format used by Zemax OpticStudio.
These values are used to parse analysis result text files.
"""

import locale
import logging

logger = logging.getLogger(__name__)

DECIMAL_POINT = locale.localeconv()["decimal_point"]
THOUSANDS_SEPARATOR = locale.localeconv()["thousands_sep"]


def set_decimal_point_and_thousands_separator() -> None:
    """Set `DECIMAL_POINT` and `THOUSHANDS_SEPARATOR` based on the system locale.

    These values are used to parse analysis result text files.

    Returns
    -------
    None

    Warnings
    --------
    This function is not thread-safe due to the use of `locale.setlocale`.
    Normal use of ZOSPy is unlikely to cause problems, but creating new threads prior to
    importing ZOSPy may cause issues. See also https://docs.python.org/3/library/locale.html#locale.setlocale
    for more information.
    """
    global DECIMAL_POINT  # noqa: PLW0603
    global THOUSANDS_SEPARATOR  # noqa: PLW0603

    old_locale = locale.setlocale(locale.LC_NUMERIC)  # get and save current numeric locale

    try:
        locale.setlocale(locale.LC_NUMERIC, "")

        DECIMAL_POINT = locale.localeconv()["decimal_point"]
        THOUSANDS_SEPARATOR = locale.localeconv()["thousands_sep"]
    except locale.Error:
        logger.error("Failed to determine decimal point and thousands separator", exc_info=True)  # noqa: G201
    finally:
        locale.setlocale(locale.LC_NUMERIC, old_locale)  # restore saved locale



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Common import ZemaxColor
from zospy.api._ZOSAPI.Editors import IEditor, IEditorCell, IEditorRow

__all__ = ("ITDERow", "IToleranceDataEditor", "ToleranceColumn", "ToleranceOperandType")

class ITDERow(IEditorRow):
    @property
    def IsActive(self) -> bool: ...
    @property
    def OperandNumber(self) -> int: ...
    @property
    def TypeName(self) -> str: ...
    @property
    def Type(self) -> ToleranceOperandType: ...
    @property
    def DoNotAdjustDuringInverseTolerancing(self) -> bool: ...
    @DoNotAdjustDuringInverseTolerancing.setter
    def DoNotAdjustDuringInverseTolerancing(self, value: bool) -> None: ...
    @property
    def IgnoreThisOperandDuringTolerancing(self) -> bool: ...
    @IgnoreThisOperandDuringTolerancing.setter
    def IgnoreThisOperandDuringTolerancing(self, value: bool) -> None: ...
    @property
    def RowColor(self) -> ZemaxColor: ...
    @RowColor.setter
    def RowColor(self, value: ZemaxColor) -> None: ...
    @property
    def Min(self) -> float: ...
    @Min.setter
    def Min(self, value: float) -> None: ...
    @property
    def MinCell(self) -> IEditorCell: ...
    @property
    def IsMinUsed(self) -> bool: ...
    @property
    def Max(self) -> float: ...
    @Max.setter
    def Max(self, value: float) -> None: ...
    @property
    def MaxCell(self) -> IEditorCell: ...
    @property
    def IsMaxUsed(self) -> bool: ...
    @property
    def Comment(self) -> str: ...
    @Comment.setter
    def Comment(self, value: str) -> None: ...
    @property
    def CommentCell(self) -> IEditorCell: ...
    @property
    def Param1(self) -> int: ...
    @Param1.setter
    def Param1(self, value: int) -> None: ...
    @property
    def Param1Cell(self) -> IEditorCell: ...
    @property
    def IsParam1Used(self) -> bool: ...
    @property
    def Param2(self) -> int: ...
    @Param2.setter
    def Param2(self, value: int) -> None: ...
    @property
    def Param2Cell(self) -> IEditorCell: ...
    @property
    def IsParam2Used(self) -> bool: ...
    @property
    def Param3(self) -> int: ...
    @Param3.setter
    def Param3(self, value: int) -> None: ...
    @property
    def Param3Cell(self) -> IEditorCell: ...
    @property
    def IsParam3Used(self) -> bool: ...
    @property
    def Nominal(self) -> float: ...
    @property
    def NominalCell(self) -> IEditorCell: ...
    @property
    def IsNominalUsed(self) -> bool: ...
    def AvailableOperandTypes(self) -> list[ToleranceOperandType]: ...
    def ChangeType(self, type: ToleranceOperandType) -> bool: ...
    def GetOperandCell(self, Col: ToleranceColumn) -> IEditorCell: ...

class IToleranceDataEditor(IEditor):
    @property
    def RowToOperandOffset(self) -> int: ...
    @property
    def NumberOfOperands(self) -> int: ...
    @property
    def FirstColumn(self) -> ToleranceColumn: ...
    @property
    def LastColumn(self) -> ToleranceColumn: ...
    @property
    def NSCToleranceWizard(self) -> INSCToleranceWizard: ...
    @property
    def SEQToleranceWizard(self) -> ISEQToleranceWizard: ...
    def AddOperand(self) -> ITDERow: ...
    def CopyOperands(self, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int) -> int: ...
    def CopyOperandsFrom(
        self, fromEditor: IToleranceDataEditor, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int
    ) -> int: ...
    def GetOperandAt(self, OperandNumber: int) -> ITDERow: ...
    def HideTDE(self) -> None: ...
    def InsertNewOperandAt(self, OperandNumber: int) -> ITDERow: ...
    def LoadToleranceFile(self, fileName: str) -> bool: ...
    def RemoveOperandAt(self, OperandNumber: int) -> bool: ...
    def RemoveOperandsAt(self, OperandNumber: int, numOperands: int) -> int: ...
    def SaveToleranceFile(self, fileName: str) -> None: ...
    def ShowTDE(self) -> bool: ...

class ToleranceColumn:
    Param1 = 2
    Param2 = 3
    Param3 = 4
    Nominal = 5
    Min = 6
    Max = 7
    Comment = 8

class ToleranceOperandType:
    TOFF = 0
    TRAD = 1
    TCUR = 2
    TFRN = 3
    TTHI = 4
    TCON = 5
    TSDI = 6
    TSDX = 7
    TSDY = 8
    TSDR = 9
    TSTX = 10
    TSTY = 11
    TIRX = 12
    TIRY = 13
    TIRR = 14
    TPAR = 15
    TEDV = 16
    TEDX = 17
    TEDY = 18
    TEDR = 19
    TETX = 20
    TETY = 21
    TETZ = 22
    TUDX = 23
    TUDY = 24
    TUTX = 25
    TUTY = 26
    TUTZ = 27
    TIND = 28
    TABB = 29
    TEXI = 30
    SAVE = 31
    COMP = 32
    CPAR = 33
    CEDV = 34
    CMCO = 35
    STAT = 36
    TWAV = 37
    TNPS = 38
    TNPA = 39
    TMCO = 40
    TEZI = 41
    TCMU = 42
    SEED = 43
    COMM = 44
    TPAI = 45
    TCIO = 46
    TCEO = 47
    TNMA = 48
    CNPA = 49
    CNPS = 50
    TRLX = 51
    TRLY = 52
    TRLR = 53
    TARX = 54
    TARY = 55
    TARR = 56



"""Old-style OpticStudio Analyses.

.. deprecated:: 2.0.0
    The `zospy.analyses.old` contains the deprecated analysis API. Migrate your analyses to the new format.

Analyses in OpticStudio are available through `zospy.analyses.old`.
These analyses are present for compatibility reasons. It is recommended to use the new analysis API.

Examples
--------
Run a Single Ray Trace analysis:

>>> from zospy.analyses.old.wavefront import zernike_standard_coefficients
>>> zernike_standard_coefficients(oss, sampling="32x32", maximum_term=15)

Open an analysis for which a wrapper function is not yet available:

>>> import zospy as zp
>>> analysis = zp.analyses.new_analysis(
...     oss, zp.constants.Analysis.AnalysisIDM.ImageSimulation
... )
"""

from warnings import warn

from zospy.analyses.old import (
    extendedscene,
    mtf,
    physicaloptics,
    polarization,
    psf,
    raysandspots,
    reports,
    surface,
    systemviewers,
    wavefront,
)
from zospy.analyses.old.base import OnComplete, new_analysis

__all__ = (
    "extendedscene",
    "mtf",
    "physicaloptics",
    "psf",
    "reports",
    "raysandspots",
    "polarization",
    "surface",
    "systemviewers",
    "wavefront",
    "new_analysis",
    "OnComplete",
)

warn("The `zospy.analyses.old` module contains the deprecated analysis API. Migrate your analyses to the new format.", DeprecationWarning)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import GiaShowAsTypes, SampleSizes
from zospy.api._ZOSAPI.Analysis.Settings import IAS_, Parity, ReferenceGia, Rotations, ScanTypes, SourceGia

__all__ = (
    "IAS_GeometricImageAnalysis",
    "IAS_ImageSimulation",
    "IAS_RelativeIllumination",
    "ISAberrationTypes",
    "ISFlipTypes",
    "ISSamplings",
    "ISShowAsTypes",
)

class IAS_GeometricImageAnalysis(IAS_):
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def Surface(self) -> IAS_Surface: ...
    @property
    def FieldSize(self) -> float: ...
    @FieldSize.setter
    def FieldSize(self, value: float) -> None: ...
    @property
    def ImageSize(self) -> float: ...
    @ImageSize.setter
    def ImageSize(self, value: float) -> None: ...
    @property
    def File(self) -> str: ...
    @File.setter
    def File(self, value: str) -> None: ...
    @property
    def Rotation(self) -> float: ...
    @Rotation.setter
    def Rotation(self, value: float) -> None: ...
    @property
    def RaysX1000(self) -> int: ...
    @RaysX1000.setter
    def RaysX1000(self, value: int) -> None: ...
    @property
    def ShowAs(self) -> GiaShowAsTypes: ...
    @ShowAs.setter
    def ShowAs(self, value: GiaShowAsTypes) -> None: ...
    @property
    def Source(self) -> SourceGia: ...
    @Source.setter
    def Source(self, value: SourceGia) -> None: ...
    @property
    def NumberOfPixels(self) -> int: ...
    @NumberOfPixels.setter
    def NumberOfPixels(self, value: int) -> None: ...
    @property
    def NA(self) -> float: ...
    @NA.setter
    def NA(self, value: float) -> None: ...
    @property
    def TotalWatts(self) -> float: ...
    @TotalWatts.setter
    def TotalWatts(self, value: float) -> None: ...
    @property
    def Parity(self) -> Parity: ...
    @Parity.setter
    def Parity(self, value: Parity) -> None: ...
    @property
    def Reference(self) -> ReferenceGia: ...
    @Reference.setter
    def Reference(self, value: ReferenceGia) -> None: ...
    @property
    def UseSymbols(self) -> bool: ...
    @UseSymbols.setter
    def UseSymbols(self, value: bool) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def RemoveVignettingFactors(self) -> bool: ...
    @RemoveVignettingFactors.setter
    def RemoveVignettingFactors(self, value: bool) -> None: ...
    @property
    def ScatterRays(self) -> bool: ...
    @ScatterRays.setter
    def ScatterRays(self, value: bool) -> None: ...
    @property
    def DeleteVignetted(self) -> bool: ...
    @DeleteVignetted.setter
    def DeleteVignetted(self, value: bool) -> None: ...
    @property
    def UsePixelInterpolation(self) -> bool: ...
    @UsePixelInterpolation.setter
    def UsePixelInterpolation(self, value: bool) -> None: ...
    @property
    def SaveAsBIMFile(self) -> str: ...
    @SaveAsBIMFile.setter
    def SaveAsBIMFile(self, value: str) -> None: ...
    @property
    def RowColumnNumber(self) -> int: ...
    @RowColumnNumber.setter
    def RowColumnNumber(self, value: int) -> None: ...
    def GetFileNames(self) -> list[str]: ...
    def UseColumnRowCenter(self) -> bool: ...

class IAS_ImageSimulation(IAS_):
    @property
    def InputFile(self) -> str: ...
    @InputFile.setter
    def InputFile(self, value: str) -> None: ...
    @property
    def FieldHeight(self) -> float: ...
    @FieldHeight.setter
    def FieldHeight(self, value: float) -> None: ...
    @property
    def FlipSource(self) -> ISFlipTypes: ...
    @FlipSource.setter
    def FlipSource(self, value: ISFlipTypes) -> None: ...
    @property
    def RotationSource(self) -> Rotations: ...
    @RotationSource.setter
    def RotationSource(self, value: Rotations) -> None: ...
    @property
    def Oversampling(self) -> ISSamplings: ...
    @Oversampling.setter
    def Oversampling(self, value: ISSamplings) -> None: ...
    @property
    def GuardBand(self) -> ISSamplings: ...
    @GuardBand.setter
    def GuardBand(self, value: ISSamplings) -> None: ...
    @property
    def Field(self) -> IAS_Field: ...
    @property
    def PupilSampling(self) -> SampleSizes: ...
    @PupilSampling.setter
    def PupilSampling(self, value: SampleSizes) -> None: ...
    @property
    def ImageSampling(self) -> SampleSizes: ...
    @ImageSampling.setter
    def ImageSampling(self, value: SampleSizes) -> None: ...
    @property
    def Aberrations(self) -> ISAberrationTypes: ...
    @Aberrations.setter
    def Aberrations(self, value: ISAberrationTypes) -> None: ...
    @property
    def PSFXPoints(self) -> int: ...
    @PSFXPoints.setter
    def PSFXPoints(self, value: int) -> None: ...
    @property
    def PSFYPoints(self) -> int: ...
    @PSFYPoints.setter
    def PSFYPoints(self, value: int) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def ApplyFixedApertures(self) -> bool: ...
    @ApplyFixedApertures.setter
    def ApplyFixedApertures(self, value: bool) -> None: ...
    @property
    def UseRelativeIllumination(self) -> bool: ...
    @UseRelativeIllumination.setter
    def UseRelativeIllumination(self, value: bool) -> None: ...
    @property
    def ShowAs(self) -> ISShowAsTypes: ...
    @ShowAs.setter
    def ShowAs(self, value: ISShowAsTypes) -> None: ...
    @property
    def Reference(self) -> ReferenceGia: ...
    @Reference.setter
    def Reference(self, value: ReferenceGia) -> None: ...
    @property
    def FlipImage(self) -> ISFlipTypes: ...
    @FlipImage.setter
    def FlipImage(self, value: ISFlipTypes) -> None: ...
    @property
    def PixelSize(self) -> float: ...
    @PixelSize.setter
    def PixelSize(self, value: float) -> None: ...
    @property
    def XPixels(self) -> int: ...
    @XPixels.setter
    def XPixels(self, value: int) -> None: ...
    @property
    def YPixels(self) -> int: ...
    @YPixels.setter
    def YPixels(self, value: int) -> None: ...
    @property
    def SuppressFrame(self) -> bool: ...
    @SuppressFrame.setter
    def SuppressFrame(self, value: bool) -> None: ...
    @property
    def OutputFile(self) -> str: ...
    @OutputFile.setter
    def OutputFile(self, value: str) -> None: ...
    def GetInputFileNames(self) -> list[str]: ...
    def GetWavelength123(self) -> bool: ...
    def GetWavelengthNumber(self) -> int: ...
    def GetWavelengthRGB(self) -> bool: ...
    def SetWavelength123(self) -> None: ...
    def SetWavelengthNumber(self, wlNum: int) -> None: ...
    def SetWavelengthRGB(self) -> None: ...

class IAS_RelativeIllumination(IAS_):
    @property
    def Wavelength(self) -> IAS_Wavelength: ...
    @property
    def ScanType(self) -> ScanTypes: ...
    @ScanType.setter
    def ScanType(self, value: ScanTypes) -> None: ...
    @property
    def RayDensity(self) -> int: ...
    @RayDensity.setter
    def RayDensity(self, value: int) -> None: ...
    @property
    def FieldDensity(self) -> int: ...
    @FieldDensity.setter
    def FieldDensity(self, value: int) -> None: ...
    @property
    def UsePolarization(self) -> bool: ...
    @UsePolarization.setter
    def UsePolarization(self, value: bool) -> None: ...
    @property
    def LogScale(self) -> bool: ...
    @LogScale.setter
    def LogScale(self, value: bool) -> None: ...
    @property
    def RemoveVignettingFactors(self) -> bool: ...
    @RemoveVignettingFactors.setter
    def RemoveVignettingFactors(self, value: bool) -> None: ...

class ISAberrationTypes:
    # None = 0
    Geometric = 1
    Diffraction = 2

class ISFlipTypes:
    # None = 0
    TopBottom = 1
    LeftRight = 2
    TopBottomLeftRight = 3

class ISSamplings:
    # None = 0
    X2 = 1
    X4 = 2
    X8 = 3
    X16 = 4
    X32 = 5
    X64 = 6

class ISShowAsTypes:
    SimulatedImage = 0
    SourceBitmap = 1
    PSFGrid = 2



import gc
import json
from typing import ClassVar

import numpy as np
import pytest
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from pydantic import TypeAdapter, ValidationError

from zospy.analyses.parsers.types import ValidatedDataFrame, ValidatedNDArray, ZOSAPIConstantAnnotation
from zospy.api import constants

validated_dataframe = TypeAdapter(ValidatedDataFrame)
validated_ndarray = TypeAdapter(ValidatedNDArray)


class TestValidatedDataFrame:
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

    @staticmethod
    def dataframe_to_json(df):
        return json.dumps(df.to_dict(orient="tight"))

    def test_validated_dataframe_to_json(self):
        json_data = validated_dataframe.dump_json(self.df).decode("utf-8")

        assert json.loads(json_data) == json.loads(self.dataframe_to_json(self.df))

    def test_validated_dataframe_from_dict(self):
        dictionary = self.df.to_dict(orient="tight")
        df = validated_dataframe.validate_python(dictionary)

        assert_frame_equal(df, self.df)

    def test_validated_dataframe_from_json(self):
        json_data = self.dataframe_to_json(self.df)
        df = validated_dataframe.validate_json(json_data)

        assert_frame_equal(df, self.df)

    def test_validated_dataframe_invalid_dict(self):
        with pytest.raises(ValidationError, match="type=invalid_dataframe"):
            # Dictionary is formatted incorrectly
            validated_dataframe.validate_python({"a": [1, 2, 3], "b": [4, 5]})

    def test_validated_dataframe_invalid_json(self):
        with pytest.raises(ValidationError, match="type=invalid_dataframe"):
            # JSON is formatted incorrectly
            validated_dataframe.validate_json(self.df.to_json(orient="split"))

    def test_validated_dataframe_invalid_type(self):
        with pytest.raises(ValidationError):
            validated_dataframe.validate_python("not a dataframe or dictionary")


class TestValidatedNDArray:
    array = np.array([[1, 2, 3], [4, 5, 6]])

    @staticmethod
    def array_to_json(array):
        return json.dumps(array.tolist())

    def test_validated_ndarray_to_json(self):
        json_data = validated_ndarray.dump_json(self.array).decode("utf-8")

        assert json.loads(json_data) == json.loads(self.array_to_json(self.array))

    def test_validated_ndarray_from_list(self):
        array = validated_ndarray.validate_python(self.array.tolist())

        assert np.array_equal(array, self.array)

    def test_validated_ndarray_from_json(self):
        json_data = self.array_to_json(self.array)
        array = validated_ndarray.validate_json(json_data)

        assert np.array_equal(array, self.array)

    def test_validated_ndarray_invalid_list(self):
        with pytest.raises(ValidationError, match="type=invalid_ndarray"):
            # List is formatted incorrectly
            validated_ndarray.validate_python([[1, 2, 3], [4, 5]])


def _get_zosapi_constant_instances():
    return [obj for obj in gc.get_objects() if isinstance(obj, ZOSAPIConstantAnnotation)]


class TestZOSAPIConstant:
    CONSTANT_MINIMUM_ZOS_VERSIONS: ClassVar[dict] = {
        "Tools": {
            "Layouts": "24.1.0",
        }
    }

    @staticmethod
    def _hasattr(obj, attr):
        for name in attr.split("."):
            if not hasattr(obj, name):
                return False

            obj = getattr(obj, name)

        return True

    def _constant_exists_in_version(self, constant: str, version) -> bool:
        _minimum_zos_version = self.CONSTANT_MINIMUM_ZOS_VERSIONS

        for namespace in constant.split("."):
            _minimum_zos_version = _minimum_zos_version.get(namespace)

            if _minimum_zos_version is None:
                return True
            if isinstance(_minimum_zos_version, str):
                return version >= _minimum_zos_version

        return True

    @pytest.mark.parametrize("annotation", _get_zosapi_constant_instances())
    def test_constant_exists(self, zos, annotation, optic_studio_version):  # noqa: ARG002
        if not self._constant_exists_in_version(annotation.enum, optic_studio_version):
            pytest.skip(f"{annotation.enum} is not available in OpticStudio {optic_studio_version}")

        assert self._hasattr(constants, annotation.enum)



"""Custom object conversions between the CLR and Python.

Python.NET allows to customize the conversion of objects between the Common Language Runtime (CLR) and Python using
Codecs [1]. This module implements codecs for automatic conversions for certain ZOS-API objects.

References
----------
.. [1] Codecs. https://pythonnet.github.io/pythonnet/codecs.html
"""

from __future__ import annotations

import logging

import clr  # noqa
import Python.Runtime
import Python.Runtime.Codecs as Codecs  # noqa

logger = logging.getLogger(__name__)


class OpticStudioInterfaceEncoder(Codecs.RawProxyEncoder):
    """Automatic downcasting of generic interfaces to specific interfaces.

    Some parts of the ZOS-API return a generic interface where a specific implementation is needed. For example,
    the settings object for an analysis can be accessed using `ZOSAPI.Analysis.IA_.GetSettings`, but the returned
    object does not give access to analysis-specific settings. In order to access these settings, the specific
    implementation needs to be used. Python.NET makes this implementation available under the `__implementation__`
    attribute. This encoder automatically downcasts certain objects to their specific implementation.

    More information about this problem is available on the `OpticStudio forum
    <https://community.zemax.com/zos-api-12/pythonnet-3-x-is-fixed-3945?tid=3945&fid=12>`.

    Only a limited number of interfaces is automatically downcast. These are documented at
    https://zospy.readthedocs.io/codecs. It is possible to register additional interfaces with
    `OpticStudioInterfaceEncoder.register_interfaces`. This should ideally be done before loading the ZOS-API with
    `zospy.zpcore.ZOS`, and in any case before the interface is used for the first time.
    """

    # The encoder will not be registered if no namespace is present
    __namespace__ = "ZOSPy.API.Codecs"

    # Types that are downcast by this encoder
    _interfaces = frozenset(
        (
            "ZOSAPI.Analysis.Settings.IAS_",
            "ZOSAPI.Editors.LDE.ISurface",
            "ZOSAPI.Editors.LDE.ISurfaceApertureType",
            "ZOSAPI.Editors.LDE.ISurfaceScatteringType",
            "ZOSAPI.Editors.NCE.IObject",
            "ZOSAPI.Tools.ISystemTool",
        )
    )

    def CanEncode(self, clr_type) -> bool:  # noqa: N802
        """Check if `clr_type` should be encoded.

        An object can be encoded if it is an interface and its full name is present in
        `OpticStudioInterfaceEncoder._interfaces`.
        """
        if clr_type.IsInterface and clr_type.FullName in self._interfaces:
            logger.debug("%s can be encoded by OpticStudioInterfaceEncoder", clr_type.FullName)
            return True

        logger.debug("%s cannot be encoded by OpticStudioInterfaceEncoder", clr_type.FullName)
        return False

    def TryEncode(self, obj):  # noqa: N802
        """Try to downcast `obj` to its implementation."""
        logger.debug("Converting %s to its implementation", obj.GetType().FullName)
        return obj.__implementation__

    @classmethod
    def register_interfaces(cls, new_interfaces: list[str] | str) -> None:
        """Register new interfaces for automatic downcasting.

        Interfaces should be registered before initializing the ZOS-API.

        Parameters
        ----------
        new_interfaces : list[str] | str
            Full interface name or list of full interface names to register.

        Examples
        --------
        >>> import zospy.api.codecs as codecs
        >>> codecs.OpticStudioInterfaceEncoder.register_interfaces(
        ...     "ZOSAPI.Editors.NCE.ISourceColorSettings"
        ... )
        """
        if isinstance(new_interfaces, str):
            new_interfaces = [new_interfaces]

        cls._interfaces |= set(new_interfaces)


def try_register_encoder(encoder: type[Python.Runtime.IPyObjectEncoder]) -> None:
    """Register `encoder` as Python.NET codec.

    Parameters
    ----------
    encoder : Type[Python.Runtime.IPyObjectEncoder]
        A class implementing `Python.Runtime.IPyObjectEncoder`. The most viable way to do this from Python is
        creating a subclass of `Python.Runtime.Codecs.RawProxyEncoder`.

    Raises
    ------
    TypeError
        If the encoder cannot be registered because it is not of the correct type.
    """
    try:
        Python.Runtime.PyObjectConversions.RegisterEncoder(encoder())
        logger.debug("Registered encoder: %s", encoder.__name__)
    except TypeError as e:
        message = f"Could not register encoder {encoder.__name__}."
        raise TypeError(message) from e


def register_codecs():
    """Register all custom codecs.

    This function should be called prior to loading the ZOS-API.
    """
    encoders = [OpticStudioInterfaceEncoder]

    for encoder in encoders:
        try_register_encoder(encoder)



"""Utility functions for ZOSPy."""

from __future__ import annotations

import re
from collections.abc import MutableMapping
from typing import Literal, TypeVar

import numpy as np
import pandas as pd

from zospy.api import _ZOSAPI
from zospy.api import config as _config


def flatten_dict(unflattened_dict, parent_key="", sep=".", *, keep_unflattened=False):
    """Flatten a nested dictionary to a single-level dictionary.

    Parameters
    ----------
    unflattened_dict : dict or MutableMapping
        A dictionary with nested dictionaries
    parent_key : str
        A key to which the flattened keys are added. Mainly present to support flatting of subdicts, should be '' in
        most primary calls.. Defaults to ''.
    sep : str
        The separator used for the flat items. Defaults to '.'
    keep_unflattened : bool
        Whether the keys, value pairs of nested dicts or MutableMappings should remain present in the output or not.
        Defaults to False

    Returns
    -------
    dict
        A dictionary with one (flattened) key for a value.
    """
    items = []
    for key, value in unflattened_dict.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, MutableMapping):
            if keep_unflattened:
                items.append((new_key, value))
            items.extend(flatten_dict(value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)


def flatten_dlltreedict(dlltreedict):
    """Flattens the set of nested dictionaries that represent a dll tree.

    For instance, it can flatten the dictionary tree obtained from ZOSAPINestedNameSpaces.json.

    Parameters
    ----------
    dlltreedict : dict
        A dictionary with nested dictionaries

    Returns
    -------
    list:
        A list with the items of the nested dictionaries.
    """
    flatdict = flatten_dict(dlltreedict, parent_key="", sep=".", keep_unflattened=True)
    return list(flatdict.keys())


def unpack_dataseries(dataseries: _ZOSAPI.Analysis.Data.IAR_DataSeries) -> pd.DataFrame:
    """Unpacks an OpticStudio dataseries in a dataframe.

    Parameters
    ----------
    dataseries : ZOSAPI.Analysis.Data.IAR_DataSeries
        Opticstudio DataSeries object.

    Returns
    -------
    pd.DataFrame
        A DataFrame holding all data and labels.
    """
    columns = pd.MultiIndex.from_product(
        [[dataseries.Description], list(dataseries.SeriesLabels)], names=["Description", "SeriesLabels"]
    )
    index = np.array(list(dataseries.XData.Data))
    data = np.array(list(dataseries.YData.Data)).reshape(dataseries.YData.Rows, dataseries.YData.Cols)
    df = pd.DataFrame(columns=columns, index=index, data=data)  # TODO evaluate
    df.index.name = dataseries.XLabel

    return df


def unpack_datagrid(
    datagrid: _ZOSAPI.Analysis.Data.IAR_DataGrid,
    minx: float | None = None,
    miny: float | None = None,
    cell_origin: Literal["bottom_left", "center"] = "bottom_left",
    label_rounding: int | None = 10,
) -> pd.DataFrame:
    """Unpack an OpticStudio datagrid to a Pandas DataFrame.

    Parameters
    ----------
    datagrid : ZOSAPI.Analysis.Data.IAR_DataGrid
        OpticStudio DataGrid object.
    minx : float, optional
        The MinX coordinate to be used when unpacking the datagrid.
    miny : float, optional
        The MinY coordinate to be used when unpacking the datagrid.
    cell_origin : Literal["bottom_left", "center"]
        Defines how minx and miny are handled to determine coordinates. Either 'bottom_left' indicating that they are
        defining the bottom left of the grd cell, or 'center', indicating that they provide the center of the grid cell.
        Defaults to 'bottom_left'.
    label_rounding : int, optional
        Defines the numbers of decimals to which the column and index labels are rounded, to fix floating point errors.
        If set to None, no rounding is applied. Defaults to 10.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the grid and the spacing on the rows and columns.
        The X- and YLabels are assigned to `df.columns.name` and `df.index.name`.
    """
    values = np.array(datagrid.Values)

    minx = datagrid.MinX if minx is None else minx
    miny = datagrid.MinY if miny is None else miny

    if cell_origin == "bottom_left":  # datagrid.MinX and .MinY point to edge of pixel, shift by half Dx and Dy
        minx += 0.5 * datagrid.Dx
        miny += 0.5 * datagrid.Dy
    elif cell_origin == "center":
        pass  # minx and miny remain equal
    else:
        raise ValueError(f"Cannot process the cell origin '{cell_origin}'")

    columns = np.linspace(minx, minx + datagrid.Dx * (datagrid.Nx - 1), datagrid.Nx)
    rows = np.linspace(miny, miny + datagrid.Dy * (datagrid.Ny - 1), datagrid.Ny)

    if label_rounding is not None:
        columns = columns.round(label_rounding)
        rows = rows.round(label_rounding)

    df = pd.DataFrame(data=values, index=rows, columns=columns)
    df.index.name = datagrid.YLabel or "y"
    df.columns.name = datagrid.XLabel or "x"

    return df


SamplingType = TypeVar("SamplingType", int, str)


def standardize_sampling(sampling: SamplingType) -> SamplingType:
    """Standardizes the sampling patterns to either int or string (S_00x00) representation.

    Parameters
    ----------
    sampling : int | str
        The sampling pattern to use. Should be int or string. Accepts both S_00x00 and 00x00 string representation.

    Returns
    -------
    int | str
        The standardized sampling pattern that can be used for processing.

    Raises
    ------
    ValueError
        When `sampling` is a string that cannot be interpreted.
    TypeError
        When `sampling` is not int or string.
    """
    if isinstance(sampling, int):
        return sampling
    if isinstance(sampling, str):
        res = re.search(r"\d+x\d+", sampling)
        if res:
            return f"S_{res.group()}"
        raise ValueError(f'Cannot interpret sampling pattern "{sampling}"')
    raise TypeError("sampling should be int or string")


# TODO: Remove in ZOSPy 2.0.0
def _get_number_field(name: str, text: str) -> str:
    return re.search(
        rf"{re.escape(name)}\s*:\s*([-+]?(\d+({re.escape(_config.DECIMAL_POINT)}\d*)?|{re.escape(_config.DECIMAL_POINT)}\d+)(?:[Ee][-+]?\d+)?)",
        text,
    ).group(1)



"""Huygens PSF analysis."""

from __future__ import annotations

from typing import Annotated, Literal, Union

from pandas import DataFrame
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import FieldNumber, WavelengthNumber  # noqa: TCH001
from zospy.api import constants
from zospy.utils.zputils import standardize_sampling

__all__ = ("HuygensPSF", "HuygensPSFSettings")


@analysis_settings
class HuygensPSFSettings:
    """Settings for the Huygens PSF analysis.

    Attributes
    ----------
    pupil_sampling : str | int
        The pupil sampling, either string (e.g. '64x64') or int. The integer will be treated as a ZOSAPI Constants
        integer.
    image_sampling : str | int
        The image sampling, either string (e.g. '64x64') or int. The integer will be treated as a ZOSAPI Constants
        integer.
    image_delta : float | int
        The image delta
    rotation : int
        The rotation, should be one of [0, 90, 180, 270].
    wavelength : str | int
        The wavelength number that is to be used. Either 'All' or an integer specifying the wavelength number.
        Defaults to 'All'.
    field : str | int
        The field number that is to be used. Either 'All' or an integer specifying the field number. Defaults to 1.
    psf_type : str
        The PSF type (e.g. 'Linear') that is calculated. Defaults to 'Linear'.
    show_as : str
        Defines how the data is shown within OpticStudio. Defaults to 'Surface'
    use_polarization : bool
        Defines if polarization is used. Defaults to `False`.
    use_centroid : bool
        Defines if centroid is used. Defaults to `False`.
    normalize : bool
        Defines if normalization is used. Defaults to `False`.
    """

    pupil_sampling: str | Annotated[int, Field(ge=0)] = Field(default="32x32", description="Pupil Sampling")
    image_sampling: str | Annotated[int, Field(ge=0)] = Field(default="32x32", description="Image Sampling")
    image_delta: float = Field(default=0, description="Image Delta")
    rotation: Literal[0, 90, 180, 270] = Field(default=0, description="Rotation")
    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number or 'All'")
    field: FieldNumber = Field(default=1, description="Field number or 'All'")
    psf_type: str = Field(default="Linear", description="PSF type")
    show_as: str = Field(default="Surface", description="Show as")
    use_polarization: bool = Field(default=False, description="Use polarization")
    use_centroid: bool = Field(default=False, description="Use centroid")
    normalize: bool = Field(default=False, description="Normalize")


class HuygensPSF(BaseAnalysisWrapper[Union[DataFrame, None], HuygensPSFSettings], analysis_type="HuygensPsf"):
    """Huygens Point Spread Function (PSF) analysis."""

    def __init__(
        self,
        *,
        pupil_sampling: str | int = "32x32",
        image_sampling: str | int = "32x32",
        image_delta: float = 0,
        rotation: int = 0,
        wavelength: str | int = "All",
        field: str | int = 1,
        psf_type: str | constants.Analysis.Settings.HuygensPsfTypes = "Linear",
        show_as: str | constants.Analysis.HuygensShowAsTypes = "Surface",
        use_polarization: bool = False,
        use_centroid: bool = False,
        normalize: bool = False,
    ):
        """Create a new Huygens PSF analysis.

        See Also
        --------
        HuygensPSFSettings : Settings for the Huygens PSF analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> DataFrame | None:
        """Run the Huygens PSF analysis."""
        self.analysis.Settings.PupilSampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.pupil_sampling)
        )
        self.analysis.Settings.ImageSampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.image_sampling)
        )
        self.analysis.Settings.ImageDelta = self.settings.image_delta
        self.analysis.Settings.Rotation = getattr(
            constants.Analysis.Settings.Rotations, f"Rotate_{self.settings.rotation}"
        )
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.Type = constants.process_constant(
            constants.Analysis.Settings.HuygensPsfTypes, self.settings.psf_type
        )
        self.analysis.Settings.ShowAsType = constants.process_constant(
            constants.Analysis.HuygensShowAsTypes, self.settings.show_as
        )
        self.analysis.Settings.UsePolarization = self.settings.use_polarization
        self.analysis.Settings.UseCentroid = self.settings.use_centroid
        self.analysis.Settings.Normalize = self.settings.normalize

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        return self.get_data_grid(cell_origin="bottom_left")



import pytest
from pandas.testing import assert_frame_equal

from zospy.analyses.psf import HuygensPSF


class TestHuygensPSF:
    def test_can_run(self, simple_system):
        result = HuygensPSF().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = HuygensPSF().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,psf_type,normalize",
        [
            ("64x64", "64x64", 0.0, "Linear", False),
            ("32x32", "64x64", 1.0, "Linear", False),
            ("128x128", "128x128", 0.0, "Real", True),
            ("32x32", "32x32", 0.0, "Real", True),
        ],
    )
    def test_huygens_psf_returns_correct_result(
        self, simple_system, pupil_sampling, image_sampling, image_delta, psf_type, normalize, expected_data
    ):
        result = HuygensPSF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            psf_type=psf_type,
            normalize=normalize,
        ).run(simple_system)

        assert_frame_equal(result.data, expected_data.data)

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,psf_type,normalize",
        [
            ("64x64", "64x64", 0.0, "Linear", False),
            ("32x32", "64x64", 1.0, "Linear", False),
            ("128x128", "128x128", 0.0, "Real", True),
            ("32x32", "32x32", 0.0, "Real", True),
        ],
    )
    def test_huygens_psf_matches_reference_data(
        self, simple_system, pupil_sampling, image_sampling, image_delta, psf_type, normalize, reference_data
    ):
        result = HuygensPSF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            psf_type=psf_type,
            normalize=normalize,
        ).run(simple_system)

        assert_frame_equal(result.data, reference_data.data)

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,psf_type,normalize",
        [
            ("64x64", "64x64", 0.0, "Linear", False),
            ("32x32", "64x64", 1.0, "Linear", False),
            ("128x128", "128x128", 0.0, "Real", True),
            ("32x32", "32x32", 0.0, "Real", True),
        ],
    )
    def test_huygens_psf_asymmetric_returns_correct_result(
        self, decentered_system, pupil_sampling, image_sampling, image_delta, psf_type, normalize, expected_data
    ):
        result = HuygensPSF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            psf_type=psf_type,
            normalize=normalize,
        ).run(decentered_system)

        assert_frame_equal(result.data, expected_data.data)

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,psf_type,normalize",
        [
            ("64x64", "64x64", 0.0, "Linear", False),
            ("32x32", "64x64", 1.0, "Linear", False),
            ("128x128", "128x128", 0.0, "Real", True),
            ("32x32", "32x32", 0.0, "Real", True),
        ],
    )
    def test_huygens_psf_asymmetric_matches_reference_data(
        self, decentered_system, pupil_sampling, image_sampling, image_delta, psf_type, normalize, reference_data
    ):
        result = HuygensPSF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            psf_type=psf_type,
            normalize=normalize,
        ).run(decentered_system)

        assert_frame_equal(result.data, reference_data.data)



"""Ray Fan (Ray Aberration) analysis."""

from __future__ import annotations

import re
from typing import Annotated, Literal

import numpy as np
import pandas as pd
from pandas import DataFrame
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.types import FieldNumber, UnitField, ValidatedDataFrame, WavelengthNumber
from zospy.api import config, constants
from zospy.utils.pyutils import atox

__all__ = ("RayFan", "RayFanSettings")


@analysis_result
class FanData:
    field_number: Annotated[int, Field(ge=0)]
    field_coordinate: UnitField[tuple[float, float]]
    data: ValidatedDataFrame

    def to_dataframe(self) -> DataFrame:
        df: DataFrame = self.data.copy()

        # Stack different columns for each wavelength
        stacked_df = pd.DataFrame(df.stack(future_stack=True), columns=["Aberration"]).reset_index(
            level=1, names=["Pupil", "Wavelength"]
        )

        stacked_df.insert(0, "Field Number", self.field_number)
        stacked_df.insert(1, "FieldX", self.field_coordinate.value[0])
        stacked_df.insert(2, "FieldY", self.field_coordinate.value[1])

        return stacked_df


@analysis_result
class RayFanResult:
    """Data for the Ray Fan analysis.

    Methods
    -------
    to_dataframe()
        Convert the data to a Pandas DataFrame.
    """

    tangential: list[FanData]
    sagittal: list[FanData]

    def to_dataframe(self) -> DataFrame:
        """Convert the data to a Pandas DataFrame.

        The separate dataframes for each field are combined in a DataFrame in long format.
        In addition to the columns for each wavelength, the returned DataFrame has the following columns:

        - Direction: The direction of the fan, either 'Tangential' or 'Sagittal'.
        - Field Number: The field number.
        - Field: The field coordinate.

        Returns
        -------
        DataFrame
            The data in long format.
        """
        dataframes = []

        for direction, fans in [("Tangential", self.tangential), ("Sagittal", self.sagittal)]:
            for fan in fans:
                df = fan.to_dataframe()
                df.insert(0, "Direction", direction)
                dataframes.append(df.reset_index())  # Move the index to a separate column to prevent overlap

        return pd.concat(dataframes, ignore_index=True)


@analysis_settings
class RayFanSettings:
    """Settings for the Ray Fan analysis.

    Attributes
    ----------
    plot_scale : float
        Maximum vertical scale for the plots. When 0, automatic scaling is used. Defaults to 0.
    number_of_rays : int
        Number of rays traced on each side of the origin of the plot.
    field : str or int
        The field number that is to be used. Either 'All' or an integer specifying the field number. Defaults to 'All'.
    wavelength : str | int
        The wavelength number that is to be used. Either 'All' or an integer specifying the wavelength number.
        Defaults to 'All'.
    tangential : str | int
        The aberration component that is plotted for the tangential fan, either 'Aberration_Y' or
        'Aberration_X'. Defaults to 'Aberration_Y'.
    sagittal : str
        The aberration component that is plotted for the sagittal fan, either 'Aberration_X' or
        'Aberration_Y'. Defaults to 'Aberration_X'.
    surface : str | int
        The surface that is to be analyzed. Either 'Image', 'Object' or an integer. Defaults to 'Image'.
    use_dashes : bool
        Defines whether solid lines or dashes are used to differentiate curves. Defaults to `False`.
    vignetted_pupil : bool
        Defines whether the pupil axis is scaled to the unvignetted pupil or not. Defaults to `True`.
    check_apertures : bool
        Defines whether only rays that pass all surface apertures are drawn or not. Defaults to `True`.
    """

    plot_scale: float = Field(default=0, ge=0, description="Plot scale, use 0 for automatic scaling")
    number_of_rays: int = Field(default=10, ge=0, description="Number of rays to trace")
    field: FieldNumber = Field(default="All", description="Field number or 'All'")
    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number or 'All'")
    tangential: str = Field(default="Aberration_Y", description="Tangential aberration to plot")
    sagittal: str = Field(default="Aberration_X", description="Sagittal aberration to plot")
    surface: Literal["Image"] | Annotated[int, Field(ge=0)] = Field(
        default="Image", description="Surface " "to be analyzed"
    )
    use_dashes: bool = Field(default=False, description="Use dashed lines for the rays")
    vignetted_pupil: bool = Field(default=True, description="Scale the pupil axis to the unvignetted pupil")
    check_apertures: bool = Field(default=True, description="Only draw rays that pass all surface apertures")


class RayFan(BaseAnalysisWrapper[RayFanResult, RayFanSettings], analysis_type="RayFan"):
    """Ray Fan analysis."""

    def __init__(
        self,
        *,
        plot_scale: float = 0,
        number_of_rays: int = 10,
        field: int | Literal["All"] = "All",
        wavelength: int | Literal["All"] = "All",
        tangential: str | constants.Analysis.Settings.Fans.TangentialAberrationComponent = "Aberration_Y",
        sagittal: str | constants.Analysis.Settings.Fans.SagittalAberrationComponent = "Aberration_X",
        surface: str | int = "Image",
        use_dashes: bool = False,
        vignetted_pupil: bool = True,
        check_apertures: bool = True,
    ):
        """Create a new Ray Fan analysis.

        See Also
        --------
        RayFanSettings : Settings for the Ray Fan analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> RayFanResult:
        """Run the Ray Fan analysis."""
        self.analysis.set_field(self.settings.field)
        self.analysis.set_surface(self.settings.surface)
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.Settings.PlotScale = self.settings.plot_scale
        self.analysis.Settings.NumberOfRays = self.settings.number_of_rays
        self.analysis.Settings.Tangential = constants.process_constant(
            constants.Analysis.Settings.Fans.TangentialAberrationComponent, self.settings.tangential
        )
        self.analysis.Settings.Sagittal = constants.process_constant(
            constants.Analysis.Settings.Fans.SagittalAberrationComponent, self.settings.sagittal
        )
        self.analysis.Settings.UseDashes = self.settings.use_dashes
        self.analysis.Settings.VignettedPupil = self.settings.vignetted_pupil
        self.analysis.Settings.CheckApertures = self.settings.check_apertures

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.get_data_series()

    def get_data_series(self) -> RayFanResult | None:
        """Get the data series from the Ray Fan analysis."""
        re_float = rf"-?\d+\{config.DECIMAL_POINT}\d+"
        ray_fan_description_regex = re.compile(
            rf"(?P<direction>sagittal|tangential) fan, field number (?P<field>\d+) = "
            rf"(?P<field_1>{re_float})(?:, (?P<field_2>{re_float}))? "
            r"\((?P<unit>.+)\)",
            re.IGNORECASE,
        )

        if self.analysis.Results.NumberOfDataSeries <= 0:
            return None

        tangential_fans = []
        sagittal_fans = []

        for i in range(self.analysis.Results.NumberOfDataSeries):
            data_series = self.analysis.Results.GetDataSeries(i)

            match = ray_fan_description_regex.match(data_series.Description)

            if match is None:
                raise ValueError(f"Could not parse description: {data_series.Description}")

            index = pd.Index(data_series.XData.Data, name=data_series.XLabel)
            columns = [atox(label, float) for label in data_series.SeriesLabels]
            data = np.array(data_series.YData.Data)

            if match.group("field_2"):
                field_x = atox(match.group("field_1"), float)
                field_y = atox(match.group("field_2"), float)
            else:  # field 1 corresponds to y
                field_x = 0.0
                field_y = atox(match.group("field_1"), float)

            coordinate = (field_x, field_y)

            fan_data = FanData(
                field_number=int(match.group("field")),
                field_coordinate=UnitField(value=coordinate, unit=match.group("unit")),
                data=DataFrame(index=index, columns=columns, data=data),
            )

            if match.group("direction").lower() == "tangential":
                tangential_fans.append(fan_data)
            else:
                sagittal_fans.append(fan_data)

        return RayFanResult(tangential=tangential_fans, sagittal=sagittal_fans)



import numpy as np
import pytest

import zospy.api.config as _config
from zospy.analyses.old.polarization import polarization_pupil_map, transmission
from zospy.utils.zputils import _get_number_field

pytestmark = pytest.mark.old_analyses

XFAIL_REASON = "Intentionally skipped for this OpticStudio version. See https://zospy.readthedocs.io/compatibility."

_signs = ["", "+", "-"]
_int_numbers = ["1", "123"]
_float_numbers = [".1", ".123", "1.", "1.2", "1.23", "12.3"]
_decimal_separators = [",", "."]
_exponents = ["", "e1", "e123", "e+1", "e+123", "e-1", "e-123", "E1", "E123", "E+1", "E+123", "E-1", "E-123"]


class TestGetNumberField:
    @pytest.mark.parametrize("exp", _exponents)
    @pytest.mark.parametrize("number", _int_numbers)
    @pytest.mark.parametrize("sign", _signs)
    def test_parses_int(self, sign, number, exp):
        number_string = sign + number + exp
        res = _get_number_field("Test", f"Test: {number_string}")

        assert res == number_string

    @pytest.mark.parametrize("decimal_separator", _decimal_separators)
    @pytest.mark.parametrize("exp", _exponents)
    @pytest.mark.parametrize("number", _float_numbers)
    @pytest.mark.parametrize("sign", _signs)
    def test_parses_float(self, sign, number, exp, decimal_separator, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr(_config, "DECIMAL_POINT", decimal_separator)

        number_string = (sign + number + exp).replace(".", decimal_separator)

        res = _get_number_field("Test", f"Test: {number_string}")

        assert res == number_string


class TestPolarizationPupilMap:
    def test_can_run_polarization_pupil_map(self, polarized_system):
        result = polarization_pupil_map(polarized_system)

        assert result.Data is not None

    @pytest.mark.parametrize(
        "jx,jy,x_phase,y_phase,surface,sampling",
        [
            (1, 0, 0, 0, "Image", "11x11"),
            (1, 1, 0, 0, 2, "11x11"),
            (0, 1, 0, 0, "Image", "11x11"),
            (1, 1, 45, 90, "Image", "17x17"),
        ],
    )
    def test_polarization_pupil_map_returns_correct_result(
        self, polarized_system, jx, jy, x_phase, y_phase, surface, sampling, expected_data
    ):
        result = polarization_pupil_map(polarized_system, jx, jy, x_phase, y_phase, surface=surface, sampling=sampling)

        assert result.Data.Transmission == expected_data.Data.Transmission
        assert np.allclose(result.Data.Table, expected_data.Data.Table)

    @pytest.mark.parametrize(
        "jx,jy,x_phase,y_phase,surface,sampling",
        [
            pytest.param(
                1,
                0,
                0,
                0,
                "Image",
                "11x11",
                marks=pytest.mark.xfail_for_opticstudio_versions(["20.3.2"], XFAIL_REASON),
            ),
            (1, 1, 0, 0, 2, "11x11"),
            (0, 1, 0, 0, "Image", "11x11"),
            pytest.param(
                1,
                1,
                45,
                90,
                "Image",
                "17x17",
                marks=pytest.mark.xfail_for_opticstudio_versions(["20.3.2", ">=24.1.3"], XFAIL_REASON),
            ),
        ],
    )
    def test_polarization_pupil_map_matches_reference_data(
        self, polarized_system, jx, jy, x_phase, y_phase, surface, sampling, reference_data
    ):
        result = polarization_pupil_map(polarized_system, jx, jy, x_phase, y_phase, surface=surface, sampling=sampling)

        assert result.Data.Transmission == reference_data.Data.Transmission
        assert np.allclose(result.Data.Table, reference_data.Data.Table)


class TestTransmission:
    def test_can_run_transmission(self, polarized_system):
        result = transmission(polarized_system)

        assert result.Data is not None

    @pytest.mark.parametrize(
        "sampling,unpolarized,jx,jy,x_phase,y_phase",
        [
            ("32x32", False, 1, 0, 0, 0),
            ("32x32", False, 0, 1, 0, 0),
            ("32x32", False, 1, 1, 0, 0),
            ("32x32", False, 0.001, 1, 0, 0),
            ("64x64", False, 1, 1, 45, 90),
            ("64x64", True, 1, 0, 0, 0),
        ],
    )
    def test_transmission_returns_correct_result(
        self, polarized_system, sampling, unpolarized, jx, jy, x_phase, y_phase, expected_data
    ):
        result = transmission(polarized_system, sampling, unpolarized, jx, jy, x_phase, y_phase)

        assert result.Data.FieldPos == expected_data.Data.FieldPos
        assert result.Data.Wavelength == expected_data.Data.Wavelength
        assert result.Data.TotalTransmission == expected_data.Data.TotalTransmission
        assert np.allclose(result.Data.Table, expected_data.Data.Table)

    @pytest.mark.parametrize(
        "sampling,unpolarized,jx,jy,x_phase,y_phase",
        [
            ("32x32", False, 1, 0, 0, 0),
            ("32x32", False, 0, 1, 0, 0),
            ("32x32", False, 1, 1, 0, 0),
            ("32x32", False, 0.001, 1, 0, 0),
            ("64x64", False, 1, 1, 45, 90),
            ("64x64", True, 1, 0, 0, 0),
        ],
    )
    def test_transmission_matches_reference_data(
        self, polarized_system, sampling, unpolarized, jx, jy, x_phase, y_phase, reference_data
    ):
        result = transmission(polarized_system, sampling, unpolarized, jx, jy, x_phase, y_phase)

        assert result.Data.FieldPos == reference_data.Data.FieldPos
        assert result.Data.Wavelength == reference_data.Data.Wavelength
        assert result.Data.TotalTransmission == reference_data.Data.TotalTransmission
        assert np.allclose(result.Data.Table, reference_data.Data.Table)

    def test_transmission_multiple_fields_raises_notimplementederror(self, simple_system):
        simple_system.SystemData.Fields.AddField(1, 1, 1.0)

        assert simple_system.SystemData.Fields.NumberOfFields == 2

        with pytest.raises(NotImplementedError):
            transmission(simple_system)

    def test_transmission_multiple_wavelengths_raises_notimplementederror(self, simple_system):
        simple_system.SystemData.Wavelengths.AddWavelength(0.314, 1.0)

        assert simple_system.SystemData.Wavelengths.NumberOfWavelengths == 2

        with pytest.raises(NotImplementedError):
            transmission(simple_system)



"""Utility functions for the Lens Data Editor (LDE) in OpticStudio."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from warnings import warn

from zospy.api import _ZOSAPI, constants

if TYPE_CHECKING:
    from zospy.zpcore import OpticStudioSystem

__all__ = ("get_pupil", "surface_change_type", "find_surface_by_comment", "surface_change_aperturetype")


@dataclass(frozen=True)
class PupilData:
    """OpticStudio pupil data.

    Attributes
    ----------
    ApertureType : int
        The aperture type.
    ApertureValue : float
        The aperture value.
    EntrancePupilDiameter : float
        The entrance pupil diameter.
    EntrancePupilPosition : float
        The entrance pupil position.
    ExitPupilDiameter : float
        The exit pupil diameter.
    ExitPupilPosition : float
        The exit pupil position.
    ApodizationType : int
        The apodization type.
    ApodizationFactor : float
        The apodization factor.
    """

    ApertureType: int
    ApertureValue: float
    EntrancePupilDiameter: float
    EntrancePupilPosition: float
    ExitPupilDiameter: float
    ExitPupilPosition: float
    ApodizationType: int
    ApodizationFactor: float


def get_pupil(oss: OpticStudioSystem):
    """Obtain the pupil data from the optical system.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance.

    Returns
    -------
    PupilData
        The pupildata.

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> oss = zos.connect()
    >>> zp.functions.lde.get_pupil(oss)
    """
    return PupilData(*oss.LDE.GetPupil())


def surface_change_type(
    surface: _ZOSAPI.Editors.LDE.ILDERow, new_type: constants.Editors.LDE.SurfaceType | str, filename=None
):
    """Change the type of a surface in the Lens Data Editor.

    Parameters
    ----------
    surface : ZOSAPI.Editors.LDE.ILDERow
        The Row/Surface for which the change is to be made.
    new_type : zospy.constants.Editors.LDE.SurfaceType | str
        The new surface type, either string (e.g. 'Standard') or int. The integer will be treated as if obtained from
        zp.constants.Editors.LDE.SurfaceType.

    Returns
    -------
    None

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> oss = zos.connect()
    >>> new_surface = oss.LDE.InsertNewSurfaceAt(0)
    >>> zp.functions.lde.surface_change_type(
    ...     new_surface, zp.constants.Editors.LDE.SurfaceType.Standard
    ... )
    """
    new_type = constants.process_constant(constants.Editors.LDE.SurfaceType, new_type)

    # Apply
    new_surface_type_settings = surface.GetSurfaceTypeSettings(new_type)

    if new_surface_type_settings.RequiresFile:
        if filename is None:
            raise ValueError(f"Surface type {new_type!s} requires the specification of a filename.")
        if filename not in list(new_surface_type_settings.GetFileNames()):
            raise ValueError(
                f"Filename '{filename}' is not listed as valid filename for this surface type. The "
                f"accepted names for this surface type are: "
                f"{', '.join(list(new_surface_type_settings.GetFileNames()))}"
            )
        new_surface_type_settings.Filename = filename

    surface.ChangeType(new_surface_type_settings)


def find_surface_by_comment(
    lde: _ZOSAPI.Editors.LDE, comment: str, *, case_sensitive: bool = True
) -> list[_ZOSAPI.Editors.LDE.ILDERow]:
    """Retrieve surfaces from the Lens Data Editor that have the supplied string as Comment.

    In case of multiple matches, the surfaces are returned in ascending order.

    Parameters
    ----------
    lde : ZOSAPI.Editors.LDE
        The Lens Data Editor (LDE)
    comment : str
        String that is searched for in the Comment column of the LDE.
    case_sensitive : bool
        Flag that specifies whether the search is case-sensitive or not. Defaults to True.

    Returns
    -------
    list[ZOSAPI.Editors.LDE.ILDERow]
        A list of surfaces in the LDE that have a Comment column value which matches the supplied comment argument.

    Examples
    --------
    >>> import zospy as zp
    >>> zos = zp.ZOS()
    >>> oss = zos.connect()
    >>> lde_object_1 = oss.LDE.GetSurfaceAt(0)
    >>> lde_object_1.Comment = "aa"
    >>> lde_object_2 = oss.LDE.GetSurfaceAt(1)
    >>> lde_object_2.Comment = "bb"
    >>> lde_object_3 = oss.LDE.GetSurfaceAt(2)
    >>> lde_object_3.Comment = "aA"
    >>> zp.functions.lde.find_surface_by_comment(oss.LDE, "aa")
    """
    # Is the search case-sensitive?
    if not case_sensitive:
        # If the search is NOT case-sensitive put comment argument in all small letters
        comment = comment.lower()

    # Initialize list of return indexes corresponding to LDE rows with matched Comment column

    return_indices = []
    # Loop over the objects and check the comments

    for surface_index in range(lde.NumberOfSurfaces):
        # Retrieve current object comment
        surface = lde.GetSurfaceAt(surface_index)
        surface_comment = surface.Comment

        # Is the search case-sensitive?
        if not case_sensitive:
            # If the search is NOT case-sensitive put current object comment in all small letters
            surface_comment = surface_comment.lower()

        # If the comment matches, store the corresponding object index
        if comment == surface_comment:
            return_indices.append(surface)

    # Return list of matched inices
    return return_indices


# Define constants for the aperture settings
_APERTURE_FILE = "ApertureFile"
_APERTURE_X_DECENTER = "ApertureXDecenter"
_APERTURE_Y_DECENTER = "ApertureYDecenter"
_MAXIMUM_RADIUS = "MaximumRadius"
_MINIMUM_RADIUS = "MinimumRadius"
_NUMBER_OF_ARMS = "NumberOfArms"
_WIDTH_OF_ARMS = "WidthOfArms"
_X_HALF_WIDTH = "XHalfWidth"
_Y_HALF_WIDTH = "YHalfWidth"
_UDA_SCALE = "UDAScale"

# Define a dictionary with accepted constants
_APERTURETYPE_USED_SETTINGS = {
    "None": (),
    "CircularAperture": (_APERTURE_X_DECENTER, _APERTURE_Y_DECENTER, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "CircularObscuration": (_APERTURE_X_DECENTER, _APERTURE_Y_DECENTER, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "Spider": (_NUMBER_OF_ARMS, _WIDTH_OF_ARMS, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "RectangularAperture": (_X_HALF_WIDTH, _Y_HALF_WIDTH, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "RectangularObscuration": (_X_HALF_WIDTH, _Y_HALF_WIDTH, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "EllipticalAperture": (_X_HALF_WIDTH, _Y_HALF_WIDTH, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "EllipticalObscuration": (_X_HALF_WIDTH, _Y_HALF_WIDTH, _MAXIMUM_RADIUS, _MINIMUM_RADIUS),
    "UserAperture": (_APERTURE_FILE, _APERTURE_X_DECENTER, _APERTURE_Y_DECENTER, _UDA_SCALE),
    "UserObscuration": (_APERTURE_FILE, _APERTURE_X_DECENTER, _APERTURE_Y_DECENTER, _UDA_SCALE),
    "FloatingAperture": (),
}


def surface_change_aperturetype(
    surface: _ZOSAPI.Editors.LDE.ILDERow,
    new_type: constants.Editors.LDE.SurfaceApertureTypes | str,
    aperture_file: str | None = None,
    aperture_x_decenter: float | None = None,
    aperture_y_decenter: float | None = None,
    maximum_radius: float | None = None,
    minimum_radius: float | None = None,
    number_of_arms: int | None = None,
    uda_scale: float | None = None,
    width_of_arms: float | None = None,
    x_half_width: float | None = None,
    y_half_width: float | None = None,
) -> None:
    """Change the aperturetype of a surface in the Lens Data Editor.

    Be aware that while all aperture parameters can be specified, only the ones accepted by the new_type should be
    given, any parameter that is not None but not accepted by the new type will result in a UserWarning and will not be
    used. Keeping a parameter on None will result in the default values used by the local OpticStudio instance.

    Parameters
    ----------
    surface : ZOSAPI.Editors.LDE.ILDERow
        The Row/Surface for which the change is to be made.
    new_type : zospy.constants.Editors.LDE.SurfaceApertureTypes | str
        The new surface aperture type.
    aperture_file : str | None
        The aperture file. Defaults to None.
    aperture_x_decenter : int | None
        The x decenter of the aperture. Defaults to None.
    aperture_y_decenter : int | None
        The y decenter of the aperture. Defaults to None.
    maximum_radius : float | None
        The maximum radius. Defaults to None.
    minimum_radius : float | None
        The minimum radius. Defaults to None.
    number_of_arms : int | None
        The number of arms. Defaults to None.
    uda_scale : float | None
        The UDA scale. Defaults to None.
    width_of_arms : float | None
        The width of arms. Defaults to None.
    x_half_width : float | None
        The x half width. Defaults to None
    y_half_width : float | None
        The y half width. Defaults to None

    Returns
    -------
    None
    """
    new_type = constants.process_constant(constants.Editors.LDE.SurfaceApertureTypes, new_type)

    new_aperturetype_settings = surface.ApertureData.CreateApertureTypeSettings(new_type)

    for param, attr_name in [
        (aperture_file, _APERTURE_FILE),
        (aperture_x_decenter, _APERTURE_X_DECENTER),
        (aperture_y_decenter, _APERTURE_Y_DECENTER),
        (maximum_radius, _MAXIMUM_RADIUS),
        (minimum_radius, _MINIMUM_RADIUS),
        (number_of_arms, _NUMBER_OF_ARMS),
        (uda_scale, _UDA_SCALE),
        (width_of_arms, _WIDTH_OF_ARMS),
        (x_half_width, _X_HALF_WIDTH),
        (y_half_width, _Y_HALF_WIDTH),
    ]:
        if param is not None:
            if attr_name not in _APERTURETYPE_USED_SETTINGS[str(new_type)]:
                warn(
                    f"Aperture type {new_type!s} does not support the specification of {attr_name}. See the "
                    f"OpticStudio documentation for more information.",
                    UserWarning,
                )
            else:
                setattr(new_aperturetype_settings, attr_name, param)

    surface.ApertureData.ChangeApertureTypeSettings(new_aperturetype_settings)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations
from . import QuickYield

__all__ = ("QuickYield", "QYCompensations", "QYCompensatorStrategy", "QYPrecisions", "ShowAsISO")

class QYCompensations:
    Standard: QYCompensations = None
    High: QYCompensations = None
    VeryHigh: QYCompensations = None

class QYCompensatorStrategy:
    OptimizeAllDampedLeastSquares: QYCompensatorStrategy = None
    ParaxialFocus: QYCompensatorStrategy = None
    Ignore: QYCompensatorStrategy = None
    OptimizeAllOrthogonalDescent: QYCompensatorStrategy = None

class QYPrecisions:
    Standard: QYPrecisions = None
    High: QYPrecisions = None
    VeryHigh: QYPrecisions = None

class ShowAsISO:
    Surface: ShowAsISO = None
    Singlet: ShowAsISO = None
    Doublet: ShowAsISO = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Tools import ISystemTool

__all__ = (
    "IGlobalOptimization",
    "IHammerOptimization",
    "ILocalOptimization",
    "OptimizationAlgorithm",
    "OptimizationCycles",
    "OptimizationSaveCount",
)

class IGlobalOptimization(ISystemTool):
    @property
    def Algorithm(self) -> OptimizationAlgorithm: ...
    @Algorithm.setter
    def Algorithm(self, value: OptimizationAlgorithm) -> None: ...
    @property
    def NumberOfCores(self) -> int: ...
    @NumberOfCores.setter
    def NumberOfCores(self, value: int) -> None: ...
    @property
    def MaxCores(self) -> int: ...
    @property
    def NumberToSave(self) -> OptimizationSaveCount: ...
    @NumberToSave.setter
    def NumberToSave(self, value: OptimizationSaveCount) -> None: ...
    @property
    def Targets(self) -> int: ...
    @property
    def Variables(self) -> int: ...
    @property
    def InitialMeritFunction(self) -> float: ...
    @property
    def Systems(self) -> int: ...
    @property
    def Cycles(self) -> int: ...
    @property
    def CurrentMeritFunction01(self) -> float: ...
    @property
    def CurrentMeritFunction02(self) -> float: ...
    @property
    def CurrentMeritFunction03(self) -> float: ...
    @property
    def CurrentMeritFunction04(self) -> float: ...
    @property
    def CurrentMeritFunction05(self) -> float: ...
    @property
    def CurrentMeritFunction06(self) -> float: ...
    @property
    def CurrentMeritFunction07(self) -> float: ...
    @property
    def CurrentMeritFunction08(self) -> float: ...
    @property
    def CurrentMeritFunction09(self) -> float: ...
    @property
    def CurrentMeritFunction10(self) -> float: ...
    @property
    def IsHPCEnabled(self) -> bool: ...
    @property
    def HPCEstimatedTimeS(self) -> float: ...
    @property
    def HPCRemainingTimeS(self) -> float: ...
    @property
    def HPCTimeToStartS(self) -> float: ...
    @property
    def HPCQueuePosition(self) -> int: ...
    @property
    def HPCState(self) -> HPCRunState: ...
    @property
    def HPCHasTimeEstimate(self) -> bool: ...
    @property
    def TargetRunTimeM(self) -> float: ...
    @TargetRunTimeM.setter
    def TargetRunTimeM(self, value: float) -> None: ...
    def CurrentMeritFunction(self, N: int) -> float: ...
    def EstimateHPCTime(self) -> bool: ...

class IHammerOptimization(ISystemTool):
    @property
    def Algorithm(self) -> OptimizationAlgorithm: ...
    @Algorithm.setter
    def Algorithm(self, value: OptimizationAlgorithm) -> None: ...
    @property
    def NumberOfCores(self) -> int: ...
    @NumberOfCores.setter
    def NumberOfCores(self, value: int) -> None: ...
    @property
    def MaxCores(self) -> int: ...
    @property
    def AutomaticOptimization(self) -> bool: ...
    @AutomaticOptimization.setter
    def AutomaticOptimization(self, value: bool) -> None: ...
    @property
    def Targets(self) -> int: ...
    @property
    def Variables(self) -> int: ...
    @property
    def InitialMeritFunction(self) -> float: ...
    @property
    def CurrentMeritFunction(self) -> float: ...
    @property
    def Systems(self) -> int: ...
    @property
    def IsHPCEnabled(self) -> bool: ...
    @property
    def HPCEstimatedTimeS(self) -> float: ...
    @property
    def HPCRemainingTimeS(self) -> float: ...
    @property
    def HPCTimeToStartS(self) -> float: ...
    @property
    def HPCQueuePosition(self) -> int: ...
    @property
    def HPCState(self) -> HPCRunState: ...
    @property
    def HPCHasTimeEstimate(self) -> bool: ...
    @property
    def TargetRunTimeM(self) -> float: ...
    @TargetRunTimeM.setter
    def TargetRunTimeM(self, value: float) -> None: ...
    def EstimateHPCTime(self) -> bool: ...

class ILocalOptimization(ISystemTool):
    @property
    def Algorithm(self) -> OptimizationAlgorithm: ...
    @Algorithm.setter
    def Algorithm(self, value: OptimizationAlgorithm) -> None: ...
    @property
    def NumberOfCores(self) -> int: ...
    @NumberOfCores.setter
    def NumberOfCores(self, value: int) -> None: ...
    @property
    def MaxCores(self) -> int: ...
    @property
    def Cycles(self) -> OptimizationCycles: ...
    @Cycles.setter
    def Cycles(self, value: OptimizationCycles) -> None: ...
    @property
    def Targets(self) -> int: ...
    @property
    def Variables(self) -> int: ...
    @property
    def InitialMeritFunction(self) -> float: ...
    @property
    def CurrentMeritFunction(self) -> float: ...

class OptimizationAlgorithm:
    DampedLeastSquares = 0
    OrthogonalDescent = 1

class OptimizationCycles:
    Automatic = 0
    Fixed_1_Cycle = 1
    Fixed_5_Cycles = 2
    Fixed_10_Cycles = 3
    Fixed_50_Cycles = 4
    Infinite = 5

class OptimizationSaveCount:
    Save_10 = 0
    Save_20 = 1
    Save_30 = 2
    Save_40 = 3
    Save_50 = 4
    Save_60 = 5
    Save_70 = 6
    Save_80 = 7
    Save_90 = 8
    Save_100 = 9



"""Zemax OpticStudio System Viewers."""

from __future__ import annotations

from inspect import signature
from os import PathLike
from typing import Any, Callable
from warnings import warn

import numpy as np
from System import Array

from zospy.analyses.old.base import AnalysisResult, OnComplete, new_analysis
from zospy.api import _ZOSAPI, constants
from zospy.utils.pyutils import abspath
from zospy.zpcore import OpticStudioSystem


def _warn_specified_parameters(
    oss: OpticStudioSystem,
    variables: dict,
    function: Callable[[Any], Any],
    ignore: tuple[str, ...] = ("oss", "oncomplete"),
) -> None:
    """Check if unsupported parameters are specified and warn the user.

    For OpticStudio versions below 24R1, compare the values of a dictionary with the default values of a function,
    and warn if any are different.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance.
    variables : dict
        A dictionary of variables to compare with the default values of the function. Can be obtained using `locals()`.
    function : Callable[[Any], Any]
        The function to compare the variables with.
    ignore : tuple[str], optional
        A tuple of parameter names to ignore. Defaults to ("oss", "oncomplete").

    Examples
    --------
    >>> def test_func(a=1, b=2, c=3):
    ...     _warn_specified_parameters(oss, locals(), test_func)
    """
    if oss.ZOS.version >= (24, 1, 0):
        return

    changed_parameters = []

    for key, value in signature(function).parameters.items():
        if key in ignore:
            continue

        if variables[key] != value.default:
            changed_parameters.append(key)

    if len(changed_parameters) > 0:
        warn(
            f"Some parameters were specified but ignored, because viewer exports are only supported from OpticStudio"
            f"24R1: {', '.join(changed_parameters)}"
        )


def _close_current_tool(oss: OpticStudioSystem) -> None:
    """Close the current tool in OpticStudio."""
    if oss.Tools.CurrentTool is not None:
        oss.Tools.CurrentTool.Close()


def _validate_wavelength(oss: OpticStudioSystem, wavelength: int | str) -> int:
    if isinstance(wavelength, str):
        if wavelength == "all":
            return -1

        raise ValueError("wavelength must be an integer or 'all'.")

    if wavelength < -1 or wavelength > oss.SystemData.Wavelengths.NumberOfWavelengths:
        raise ValueError("wavelength must be -1 or between 1 and the number of wavelengths.")

    return wavelength


def _validate_field(oss: OpticStudioSystem, field: int | str) -> int:
    if isinstance(field, str):
        if field == "all":
            return -1

        raise ValueError("field must be an integer or 'all'.")

    if field < 1 or field > oss.SystemData.Fields.NumberOfFields:
        raise ValueError("field must be -1 or between 1 and the number of fields.")

    return field


ALLOWED_IMAGE_EXTENSIONS: tuple[str, ...] = ("bmp", "jpeg", "png")


def _validate_path(path: PathLike | str) -> str:
    str_path = abspath(path, check_directory_only=True)

    if str_path.split(".")[-1] in ALLOWED_IMAGE_EXTENSIONS:
        return str_path

    raise ValueError(f"Image file must have one of the following extensions: {ALLOWED_IMAGE_EXTENSIONS}")


def _get_image_data(image_data: _ZOSAPI.Tools.Layouts.IImageExportData | None) -> np.ndarray | None:
    if image_data is None:
        return image_data

    image_size = image_data.Width * image_data.Height

    # In-place updating arrays works only with dotnet arrays
    r_values = Array[int](image_size)  # [0] * image_size
    g_values = Array[int](image_size)  # [0] * image_size
    b_values = Array[int](image_size)  # [0] * image_size

    image_data.FillValues(image_data.Width * image_data.Height, r_values, g_values, b_values)

    return np.stack((r_values, g_values, b_values), axis=-1).reshape(image_data.Height, image_data.Width, 3)


def cross_section(
    oss: OpticStudioSystem,
    start_surface: int = 1,
    end_surface: int = -1,
    number_of_rays: int = 3,
    y_stretch: float = 1.0,
    fletch_rays: bool = False,
    wavelength: int | str = "all",
    field: int | str = "all",
    color_rays_by: constants.Tools.Layouts.ColorRaysByCrossSectionOptions | str = "Fields",
    upper_pupil: float = 1,
    lower_pupil: float = -1,
    delete_vignetted: bool = False,
    marginal_and_chief_only: bool = False,
    image_size: tuple[int, int] = (800, 600),
    line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
    imgoutfile: PathLike | str | None = None,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Cross-Section viewer.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    start_surface : int, optional
        The starting surface index for the cross-section analysis. Defaults to 1.
    end_surface : int, optional
        The ending surface index for the cross-section analysis. A value of -1 indicates the last surface.
        Defaults to -1.
    number_of_rays : int, optional
        The number of rays to be used in the analysis. Defaults to 3.
    y_stretch : float, optional
        The stretch factor in the Y-axis for the analysis visualization. Defaults to 1.0.
    fletch_rays : bool, optional
        Flag indicating whether to fletch rays. Defaults to False.
    wavelength : int | str, optional
        The wavelength index to be used. Can be an integer or "all" for all wavelengths. Defaults to "all".
    field : int | str, optional
        The field index to be used. Can be an integer or "all" for all fields. Defaults to "all".
    color_rays_by : constants.Tools.Layouts.ColorRaysByCrossSectionOptions | str, optional
        The criterion for coloring rays in the analysis. Defaults to "Fields".
    upper_pupil : float, optional
        The upper pupil limit for the analysis. Defaults to 1.
    lower_pupil : float, optional
        The lower pupil limit for the analysis. Defaults to -1.
    delete_vignetted : bool, optional
        Flag indicating whether to delete vignetted rays. Defaults to False.
    marginal_and_chief_only : bool, optional
        Flag indicating whether to include only marginal and chief rays in the analysis. Defaults to False.
    image_size : tuple[int, int], optional
        The size of the output image in pixels (width, height). Defaults to (800, 600).
    line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines in the output visualization. Defaults to "Standard".
    imgoutfile : PathLike | str | None, optional
        The path to save the output image file. If None, no image is saved. Defaults to None.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Cross-Section analysis result. If `imgoutfile` is `None`, the result will contain the image data as a numpy
        array. If `imgoutfile` is not `None`, the results `Data` attribute will be empty.
    """
    _warn_specified_parameters(oss, locals(), cross_section)

    if start_surface < -1:
        raise ValueError("start_surface must be greater than or equal to -1.")

    if end_surface != -1 and end_surface <= start_surface or end_surface > oss.LDE.NumberOfSurfaces - 1:
        raise ValueError("end_surface must be -1 or greater than start_surface and less than the number of surfaces.")

    if end_surface == -1:
        # According to the documentation the ZOS-API should interpret -1 as the last surface,
        # but instead it interprets it as the first surface after the start surface.
        end_surface = oss.LDE.NumberOfSurfaces - 1

    if number_of_rays < 1:
        raise ValueError("number_of_rays must be greater than 0.")

    if y_stretch < 0:
        raise ValueError("y_stretch must be greater than or equal to 0.")

    wavelength = _validate_wavelength(oss, wavelength)
    field = _validate_field(oss, field)

    analysis_type = constants.Analysis.AnalysisIDM.Draw2D

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    if oss.ZOS.version >= (24, 1, 0):
        _close_current_tool(oss)

        layout_tool = oss.Tools.Layouts.OpenCrossSectionExport()
        layout_tool.StartSurface = start_surface
        layout_tool.EndSurface = end_surface
        layout_tool.NumberOfRays = number_of_rays
        layout_tool.YStretch = y_stretch
        layout_tool.FletchRays = fletch_rays
        layout_tool.Wavelength = wavelength
        layout_tool.Field = field
        layout_tool.ColorRaysBy = constants.process_constant(
            constants.Tools.Layouts.ColorRaysByCrossSectionOptions, color_rays_by
        )
        layout_tool.UpperPupil = upper_pupil
        layout_tool.LowerPupil = lower_pupil
        layout_tool.DeleteVignetted = delete_vignetted
        layout_tool.MarginalAndChiefOnly = marginal_and_chief_only
        layout_tool.OutputPixelWidth, layout_tool.OutputPixelHeight = image_size
        layout_tool.RaysLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, line_thickness
        )

        if imgoutfile is not None:
            layout_tool.SaveImageAsFile = True
            layout_tool.OutputFileName = _validate_path(imgoutfile)
        else:
            layout_tool.SaveImageAsFile = False

        layout_tool.RunAndWaitForCompletion()

        if not layout_tool.Succeeded:
            raise RuntimeError("The cross-section export tool failed to run.")

        image_data = _get_image_data(layout_tool.ImageExportData) if imgoutfile is None else None
    else:
        image_data = None

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=image_data,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def viewer_3d(
    oss: OpticStudioSystem,
    start_surface: int = 1,
    end_surface: int = -1,
    number_of_rays: int = 3,
    wavelength: int | str = "all",
    field: int | str = "all",
    ray_pattern: constants.Tools.General.RayPatternType | str = "XYFan",
    color_rays_by: constants.Tools.Layouts.ColorRaysByOptions | str = "Fields",
    delete_vignetted: bool = False,
    hide_lens_faces: bool = False,
    hide_lens_edges: bool = False,
    hide_x_bars: bool = False,
    draw_paraxial_pupils: bool = False,
    fletch_rays: bool = False,
    split_nsc_rays: bool = False,
    scatter_nsc_rays: bool = False,
    draw_real_entrance_pupils: constants.Tools.Layouts.RealPupilOptions | str = "Pupils_Off",
    draw_real_exit_pupils: constants.Tools.Layouts.RealPupilOptions | str = "Pupils_Off",
    surface_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
    ray_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
    configuration_all: bool = False,
    configuration_current: bool = False,
    configuration_offset_x: float = 0,
    configuration_offset_y: float = 0,
    configuration_offset_z: float = 0,
    camera_viewpoint_angle_x: float = 0,
    camera_viewpoint_angle_y: float = 0,
    camera_viewpoint_angle_z: float = 0,
    imgoutfile: PathLike | str | None = None,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio 3D Viewer.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    start_surface : int, optional
        The starting surface index for the 3D viewer. Defaults to 1.
    end_surface : int, optional
        The ending surface index for the 3D viewer. A value of -1 indicates the last surface. Defaults to -1.
    number_of_rays : int, optional
        The number of rays to be used in the analysis. Defaults to 3.
    wavelength : int | str, optional
        The wavelength index to be used. Can be an integer or "all" for all wavelengths. Defaults to "all".
    field : int | str, optional
        The field index to be used. Can be an integer or "all" for all fields. Defaults to "all".
    ray_pattern : constants.Tools.General.RayPatternType | str, optional
        The ray pattern to be used in the analysis. Defaults to "XYFan".
    color_rays_by : constants.Tools.Layouts.ColorRaysByOptions | str, optional
        The criterion for coloring rays in the analysis. Defaults to "Fields".
    delete_vignetted : bool, optional
        Flag indicating whether to delete vignetted rays. Defaults to False.
    hide_lens_faces : bool, optional
        Flag indicating whether to hide lens faces. Defaults to False.
    hide_lens_edges : bool, optional
        Flag indicating whether to hide lens edges. Defaults to False.
    hide_x_bars : bool, optional
        Flag indicating whether to hide the x-component of lens faces. Defaults to False.
    draw_paraxial_pupils : bool, optional
        Flag indicating whether to draw paraxial entrance and exit pupils. Defaults to False.
    fletch_rays : bool, optional
        Flag indicating whether to draw small arrows indicating the direction of the rays. Defaults to False.
    split_nsc_rays : bool, optional
        Flag indicating whether to split rays from non-sequential sources at ray-surface intercepts. Defaults to False.
    scatter_nsc_rays : bool, optional
        Flag indicating whether to scatter rays from non-sequential sources at ray-surface intercepts.
        Defaults to False.
    draw_real_entrance_pupils : constants.Tools.Layouts.RealPupilOptions | str, optional
        How to draw real entrance pupils. Defaults to "Pupils_Off". Can be one of ['Pupils_Off', 'Pupils_4',
        'Pupils_8', 'Pupils_16', 'Pupils_32'].
    draw_real_exit_pupils : constants.Tools.Layouts.RealPupilOptions | str, optional
        How to draw real exit pupils. Defaults to "Pupils_Off". Can be one of ['Pupils_Off', 'Pupils_4', 'Pupils_8',
        'Pupils_16', 'Pupils_32'].
    surface_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines for the surfaces. Defaults to "Standard".
    ray_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines for the rays. Defaults to "Standard".
    configuration_all : bool, optional
        Flag indicating whether to use all configurations, if multiple configurations are present. Defaults to False.
        If multiple configurations are displayed, the `configuration_offset_*` parameters can be used to add offsets
        between the different configurations.
    configuration_current : bool, optional
        Flag indicating whether to only display the current configuration, when multiple configurations are present.
        Defaults to False.
    configuration_offset_x : float, optional
        The offset along the X-axis between configurations, if multiple configurations are present. Defaults to 0.
    configuration_offset_y : float, optional
        The offset along the Y-axis between configurations, if multiple configurations are present. Defaults to 0.
    configuration_offset_z : float, optional
        The offset along the Z-axis between configurations, if multiple configurations are present. Defaults to 0.
    camera_viewpoint_angle_x : float, optional
        Rotation of the system around the X-axis, in degrees. Defaults to 0.
    camera_viewpoint_angle_y : float, optional
        Rotation of the system around the Y-axis, in degrees. Defaults to 0.
    camera_viewpoint_angle_z : float, optional
        Rotation of the system around the Z-axis, in degrees. Defaults to 0.
    imgoutfile : PathLike | str | None, optional
        The path to save the output image file. If None, no image is saved. Defaults to None.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Viewer 3D analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    _warn_specified_parameters(oss, locals(), viewer_3d)

    if start_surface < -1:
        raise ValueError("start_surface must be greater than or equal to -1.")

    if end_surface != -1 and end_surface <= start_surface or end_surface > oss.LDE.NumberOfSurfaces - 1:
        raise ValueError("end_surface must be -1 or greater than start_surface and less than the number of surfaces.")

    if end_surface == -1:
        # According to the documentation the ZOS-API should interpret -1 as the last surface,
        # but instead it interprets it as the first surface after the start surface.
        end_surface = oss.LDE.NumberOfSurfaces - 1

    if number_of_rays < 1:
        raise ValueError("number_of_rays must be greater than 0.")

    analysis_type = constants.Analysis.AnalysisIDM.Draw3D

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    if oss.ZOS.version >= (24, 1, 0):
        _close_current_tool(oss)

        layout_tool = oss.Tools.Layouts.Open3DViewerExport()

        layout_tool.StartSurface = start_surface
        layout_tool.EndSurface = end_surface
        layout_tool.NumberOfRays = number_of_rays
        layout_tool.Wavelength = _validate_wavelength(oss, wavelength)
        layout_tool.Field = _validate_field(oss, field)
        layout_tool.RayPattern = constants.process_constant(constants.Tools.General.RayPatternType, ray_pattern)
        layout_tool.ColorRaysBy = constants.process_constant(constants.Tools.Layouts.ColorRaysByOptions, color_rays_by)
        layout_tool.DeleteVignetted = delete_vignetted
        layout_tool.HideLensFaces = hide_lens_faces
        layout_tool.HideLensEdges = hide_lens_edges
        layout_tool.HideXBars = hide_x_bars
        layout_tool.DrawParaxialPupils = draw_paraxial_pupils
        layout_tool.FletchRays = fletch_rays
        layout_tool.SplitNSCRays = split_nsc_rays
        layout_tool.ScatterNSCRays = scatter_nsc_rays
        layout_tool.DrawRealEntrancePupils = constants.process_constant(
            constants.Tools.Layouts.RealPupilOptions, draw_real_entrance_pupils
        )
        layout_tool.DrawRealExitPupils = constants.process_constant(
            constants.Tools.Layouts.RealPupilOptions, draw_real_exit_pupils
        )
        layout_tool.SurfaceLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, surface_line_thickness
        )
        layout_tool.RaysLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, ray_line_thickness
        )
        layout_tool.ConfigurationAll = configuration_all
        layout_tool.ConfigurationCurrent = configuration_current
        layout_tool.ConfigurationOffsetX = configuration_offset_x
        layout_tool.ConfigurationOffsetY = configuration_offset_y
        layout_tool.ConfigurationOffsetZ = configuration_offset_z
        layout_tool.CameraViewpointAngleX = camera_viewpoint_angle_x
        layout_tool.CameraViewpointAngleY = camera_viewpoint_angle_y
        layout_tool.CameraViewpointAngleZ = camera_viewpoint_angle_z

        if imgoutfile is not None:
            layout_tool.SaveImageAsFile = True
            layout_tool.OutputFileName = _validate_path(imgoutfile)
        else:
            layout_tool.SaveImageAsFile = False

        layout_tool.RunAndWaitForCompletion()

        if not layout_tool.Succeeded:
            raise RuntimeError("The 3D viewer export tool failed to run.")

        image_data = _get_image_data(layout_tool.ImageExportData) if imgoutfile is None else None
    else:
        image_data = None

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=image_data,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def shaded_model(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio Shaded Model Viewer.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be sequential.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A Shaded Model analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.ShadedModel

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def nsc_3d_layout(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio NSC 3D Layout viewer.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be non-sequential.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A NSC 3D Layout analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.NSC3DLayout

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)


def nsc_shaded_model(
    oss: OpticStudioSystem,
    oncomplete: OnComplete | str = OnComplete.Release,
) -> AnalysisResult:
    """Wrapper around the OpticStudio NSC Shaded Model Viewer.

    Parameters
    ----------
    oss : zospy.zpcore.OpticStudioSystem
        A ZOSPy OpticStudioSystem instance. Should be non-sequential.
    oncomplete : OnComplete | str
        Defines behaviour upon completion of the analysis. Should be one of ['Close', 'Release', 'Sustain']. If 'Close',
        the analysis will be closed after completion. If 'Release', the analysis will remain open in OpticStudio, but
        the link with python will be destroyed. If 'Sustain' the analysis will be kept open in OpticStudio and the link
        with python will be sustained. To enable interaction when oncomplete == 'Sustain', the OpticStudio Analysis
        instance will be available in the returned AnalysisResult through AnalysisResult.Analysis. Defaults to
        'Release'.

    Returns
    -------
    AnalysisResult
        A NSC Shaded Model analysis result. As the viewers do not return data, the AnalysisResult can mainly be used to
        further control the analysis when oncomplete is set to `Sustain`.
    """
    analysis_type = constants.Analysis.AnalysisIDM.NSCShadedModel

    analysis = new_analysis(oss, analysis_type, settings_first=False)

    # Get headerdata, metadata and messages
    headerdata = analysis.get_header_data()
    metadata = analysis.get_metadata()
    messages = analysis.get_messages()

    # Create output
    result = AnalysisResult(
        analysistype=str(analysis_type),
        data=None,
        settings=None,
        metadata=metadata,
        headerdata=headerdata,
        messages=messages,
    )

    return analysis.complete(oncomplete, result)



# /// script
# dependencies = [
#   "ipykernel",
#   "nbconvert",
# ]
# ///

"""Run a ZOSPy example.

This script is used to run a single ZOSPy example. It is called by `run_all_examples.py` or `run_single_example.py`;
these scripts make sure the notebook is run in a separate environment with the required dependencies.
The script reads a Jupyter notebook, changes the connection mode from extension to standalone
(unless configured to do so), runs the notebook, and saves the output.
"""

from __future__ import annotations

import argparse
import re
import sys
import traceback
from pathlib import Path
from typing import Literal

import nbformat
from nbconvert.preprocessors import CellExecutionError, ExecutePreprocessor

ConnectionMode = Literal["standalone", "extension"]


def change_connection_mode(
    notebook: nbformat.NotebookNode,
    from_mode: ConnectionMode = "standalone",
    to_mode: ConnectionMode = "extension",
) -> str | None:
    regex_mode = re.compile(rf'zos\.connect\(\s*(?:mode\s*=\s*)?"{from_mode}"\s*\)')

    for cell in notebook.cells:
        if cell.cell_type == "code" and regex_mode.search(cell.source):
            mode = regex_mode.search(cell.source).group(0)
            cell.source = regex_mode.sub(f'zos.connect("{to_mode}")', cell.source)
            return mode

    return None


def process_notebook(
    notebook_path: Path,
    output_folder: Path | None = None,
    *,
    keep_extension_mode: bool = False,
) -> bool:
    nb = nbformat.read(notebook_path, as_version=4)

    # Change the connection mode to 'standalone' to run the notebook
    old_connection_mode = change_connection_mode(nb, "extension", "standalone") if not keep_extension_mode else False

    # Run the notebook
    ep = ExecutePreprocessor(kernel_name="python3")
    try:
        ep.preprocess(nb, {"metadata": {"path": notebook_path.parent}})
    except CellExecutionError:
        print(f"Error executing {notebook_path}: {traceback.format_exc()}")
        return False

    # Reverse the connection mode change
    if old_connection_mode:
        change_connection_mode(nb, "standalone", "extension")

    # Save the output
    output_path = notebook_path if output_folder is None else output_folder / notebook_path.name
    print(f"Saving output to {output_path}")
    nbformat.write(nb, output_path)

    return True


def main(args: argparse.Namespace) -> int:
    example_path: Path = Path(args.example).resolve(strict=False)
    output_path: Path = args.output or example_path

    if not example_path.exists():
        print(f"Directory {example_path} does not exist")
        return -1

    if not example_path.is_dir():
        print(f"Notebook {example_path} is not a directory. Please provide a directory.")
        return -1

    if not output_path.exists():
        print(f"Creating output directory at {args.output}")
        try:
            output_path.mkdir()
        except FileNotFoundError:
            print(f"Error creating output directory: {traceback.format_exc()}")
            return -1

    exitcode = 0

    for notebook in example_path.glob("*.ipynb"):
        print(f"Processing {notebook}...")
        success = process_notebook(notebook, args.output, keep_extension_mode=args.keep_extension_mode)

        if not success:
            print(f"Failed to process {notebook}.")
            exitcode += 1

    return exitcode


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Run a single example")
    parser.add_argument("example", type=Path, help="Path to the example directory")
    parser.add_argument(
        "--output",
        type=Path,
        help="Path to save the output to. Defaults to the input example directory.",
    )
    parser.add_argument(
        "--keep-extension-mode",
        action="store_true",
        help="Keep the extension connection mode after running the notebook",
    )

    args = parser.parse_args()

    sys.exit(main(args))



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Analysis import IMessage

__all__ = ("IAS_QYField", "XYSymmetricField", "YSymmetricField")

class IAS_QYField:
    @property
    def IsFieldUser(self) -> bool: ...
    @property
    def IsFieldSymmetricXY(self) -> bool: ...
    @property
    def IsFieldSymmetricY(self) -> bool: ...
    def GetFieldSymmetricXY(self) -> XYSymmetricField: ...
    def GetFieldSymmetricY(self) -> YSymmetricField: ...
    def GetFieldUser(self) -> int: ...
    def SetFieldSymmetricXY(self, field: XYSymmetricField) -> IMessage: ...
    def SetFieldSymmetricY(self, field: YSymmetricField) -> IMessage: ...
    def SetFieldUser(self, fieldNumber: int) -> IMessage: ...

class XYSymmetricField:
    All = 0
    Zero = 1
    YPositive70Percent = 2
    YNegative70Percent = 3
    YPositive100Percent = 4
    YNegative100Percent = 5
    XPositive70Percent = 6
    XNegative70Percent = 7
    XPositive100Percent = 8
    XNegative100Percent = 9

class YSymmetricField:
    All = 0
    Zero = 1
    Positive70Percent = 2
    Negative70Percent = 3
    Positive100Percent = 4
    Negative100Percent = 5
